"""
Contains Python equivalents of the function and constant
definitions in CANLIB's canlib.h, with some supporting functionality
specific to Python.

Copyright (C) 2010 Dynamic Controls
"""

import ctypes
import logging
import sys
import time

from can import BusABC, Message
from can.util import time_perfcounter_correlation

from ...exceptions import CanError, CanInitializationError, CanOperationError
from . import constants as canstat
from . import structures

log = logging.getLogger("can.kvaser")

# Resolution in us
TIMESTAMP_RESOLUTION = 10

TIMESTAMP_FACTOR = TIMESTAMP_RESOLUTION / 1000000.0


try:
    if sys.platform == "win32":
        __canlib = ctypes.windll.LoadLibrary("canlib32")
    else:
        __canlib = ctypes.cdll.LoadLibrary("libcanlib.so")
    log.info("loaded kvaser's CAN library")
except OSError:
    log.warning("Kvaser canlib is unavailable.")
    __canlib = None


def _unimplemented_function(*args):
    raise NotImplementedError("This function is not implemented in canlib")


def __get_canlib_function(func_name, argtypes=None, restype=None, errcheck=None):
    argtypes = [] if argtypes is None else argtypes
    # log.debug('Wrapping function "%s"' % func_name)
    try:
        # e.g. canlib.canBusOn
        retval = getattr(__canlib, func_name)
        # log.debug('"%s" found in library', func_name)
    except AttributeError:
        log.warning('"%s" was not found in library', func_name)
        return _unimplemented_function
    else:
        # log.debug('Result type is: %s' % type(restype))
        # log.debug('Error check function is: %s' % errcheck)
        retval.argtypes = argtypes
        retval.restype = restype
        if errcheck:
            retval.errcheck = errcheck
        return retval


class CANLIBError(CanError):

    """
    Try to display errors that occur within the wrapped C library nicely.
    """

    def __init__(self, function, error_code, arguments):
        message = CANLIBError._get_error_message(error_code)
        super().__init__(f"Function {function.__name__} failed - {message}", error_code)
        self.function = function
        self.arguments = arguments

    @staticmethod
    def _get_error_message(error_code: int) -> str:
        errmsg = ctypes.create_string_buffer(128)
        canGetErrorText(error_code, errmsg, len(errmsg))
        return errmsg.value.decode("ascii")


class CANLIBInitializationError(CANLIBError, CanInitializationError):
    pass


class CANLIBOperationError(CANLIBError, CanOperationError):
    pass


def __convert_can_status_to_int(result):
    if isinstance(result, int):
        return result
    else:
        return result.value


def __check_status_operation(result, function, arguments):
    result = __convert_can_status_to_int(result)
    if not canstat.CANSTATUS_SUCCESS(result):
        raise CANLIBOperationError(function, result, arguments)
    return result


def __check_status_initialization(result, function, arguments):
    result = __convert_can_status_to_int(result)
    if not canstat.CANSTATUS_SUCCESS(result):
        raise CANLIBInitializationError(function, result, arguments)
    return result


def __check_status_read(result, function, arguments):
    result = __convert_can_status_to_int(result)
    if not canstat.CANSTATUS_SUCCESS(result) and result != canstat.canERR_NOMSG:
        raise CANLIBOperationError(function, result, arguments)
    return result


class c_canHandle(ctypes.c_int):
    pass


canINVALID_HANDLE = -1


def __check_bus_handle_validity(handle, function, arguments):
    if handle.value > canINVALID_HANDLE:
        return handle  # is valid

    result = __convert_can_status_to_int(handle)
    raise CANLIBInitializationError(function, result, arguments)


if __canlib is not None:
    canInitializeLibrary = __get_canlib_function("canInitializeLibrary")

    canGetErrorText = __get_canlib_function(
        "canGetErrorText",
        argtypes=[canstat.c_canStatus, ctypes.c_char_p, ctypes.c_uint],
        restype=canstat.c_canStatus,
        errcheck=__check_status_operation,
    )

    # TODO wrap this type of function to provide a more Pythonic API
    canGetNumberOfChannels = __get_canlib_function(
        "canGetNumberOfChannels",
        argtypes=[ctypes.c_void_p],
        restype=canstat.c_canStatus,
        errcheck=__check_status_initialization,
    )

    kvReadTimer = __get_canlib_function(
        "kvReadTimer",
        argtypes=[c_canHandle, ctypes.POINTER(ctypes.c_uint)],
        restype=canstat.c_canStatus,
        errcheck=__check_status_initialization,
    )

    canBusOff = __get_canlib_function(
        "canBusOff",
        argtypes=[c_canHandle],
        restype=canstat.c_canStatus,
        errcheck=__check_status_operation,
    )

    canBusOn = __get_canlib_function(
        "canBusOn",
        argtypes=[c_canHandle],
        restype=canstat.c_canStatus,
        errcheck=__check_status_initialization,
    )

    canClose = __get_canlib_function(
        "canClose",
        argtypes=[c_canHandle],
        restype=canstat.c_canStatus,
        errcheck=__check_status_operation,
    )

    canOpenChannel = __get_canlib_function(
        "canOpenChannel",
        argtypes=[ctypes.c_int, ctypes.c_int],
        restype=c_canHandle,
        errcheck=__check_bus_handle_validity,
    )

    canSetBusParams = __get_canlib_function(
        "canSetBusParams",
        argtypes=[
            c_canHandle,
            ctypes.c_long,
            ctypes.c_uint,
            ctypes.c_uint,
            ctypes.c_uint,
            ctypes.c_uint,
            ctypes.c_uint,
        ],
        restype=canstat.c_canStatus,
        errcheck=__check_status_initialization,
    )

    canSetBusParamsFd = __get_canlib_function(
        "canSetBusParamsFd",
        argtypes=[
            c_canHandle,
            ctypes.c_long,
            ctypes.c_uint,
            ctypes.c_uint,
            ctypes.c_uint,
        ],
        restype=canstat.c_canStatus,
        errcheck=__check_status_initialization,
    )

    canSetBusOutputControl = __get_canlib_function(
        "canSetBusOutputControl",
        argtypes=[c_canHandle, ctypes.c_uint],
        restype=canstat.c_canStatus,
        errcheck=__check_status_initialization,
    )

    canSetAcceptanceFilter = __get_canlib_function(
        "canSetAcceptanceFilter",
        argtypes=[c_canHandle, ctypes.c_uint, ctypes.c_uint, ctypes.c_int],
        restype=canstat.c_canStatus,
        errcheck=__check_status_operation,
    )

    canReadWait = __get_canlib_function(
        "canReadWait",
        argtypes=[
            c_canHandle,
            ctypes.c_void_p,
            ctypes.c_void_p,
            ctypes.c_void_p,
            ctypes.c_void_p,
            ctypes.c_void_p,
            ctypes.c_long,
        ],
        restype=canstat.c_canStatus,
        errcheck=__check_status_read,
    )

    canWrite = __get_canlib_function(
        "canWrite",
        argtypes=[
            c_canHandle,
            ctypes.c_long,
            ctypes.c_void_p,
            ctypes.c_uint,
            ctypes.c_uint,
        ],
        restype=canstat.c_canStatus,
        errcheck=__check_status_operation,
    )

    canWriteSync = __get_canlib_function(
        "canWriteSync",
        argtypes=[c_canHandle, ctypes.c_ulong],
        restype=canstat.c_canStatus,
        errcheck=__check_status_operation,
    )

    canIoCtlInit = __get_canlib_function(
        "canIoCtl",
        argtypes=[c_canHandle, ctypes.c_uint, ctypes.c_void_p, ctypes.c_uint],
        restype=canstat.c_canStatus,
        errcheck=__check_status_initialization,
    )

    canIoCtl = __get_canlib_function(
        "canIoCtl",
        argtypes=[c_canHandle, ctypes.c_uint, ctypes.c_void_p, ctypes.c_uint],
        restype=canstat.c_canStatus,
        errcheck=__check_status_operation,
    )

    canGetVersion = __get_canlib_function(
        "canGetVersion", restype=ctypes.c_short, errcheck=__check_status_operation
    )

    kvFlashLeds = __get_canlib_function(
        "kvFlashLeds",
        argtypes=[c_canHandle, ctypes.c_int, ctypes.c_int],
        restype=ctypes.c_short,
        errcheck=__check_status_operation,
    )

    if sys.platform == "win32":
        canGetVersionEx = __get_canlib_function(
            "canGetVersionEx",
            argtypes=[ctypes.c_uint],
            restype=ctypes.c_uint,
            errcheck=__check_status_operation,
        )

    canGetChannelData = __get_canlib_function(
        "canGetChannelData",
        argtypes=[ctypes.c_int, ctypes.c_int, ctypes.c_void_p, ctypes.c_size_t],
        restype=canstat.c_canStatus,
        errcheck=__check_status_initialization,
    )

    canRequestBusStatistics = __get_canlib_function(
        "canRequestBusStatistics",
        argtypes=[c_canHandle],
        restype=canstat.c_canStatus,
        errcheck=__check_status_operation,
    )

    canGetBusStatistics = __get_canlib_function(
        "canGetBusStatistics",
        argtypes=[
            c_canHandle,
            ctypes.POINTER(structures.BusStatistics),
            ctypes.c_size_t,
        ],
        restype=canstat.c_canStatus,
        errcheck=__check_status_operation,
    )


def init_kvaser_library():
    if __canlib is not None:
        try:
            log.debug("Initializing Kvaser CAN library")
            canInitializeLibrary()
            log.debug("CAN library initialized")
        except Exception:
            log.warning("Kvaser canlib could not be initialized.")


DRIVER_MODE_SILENT = False
DRIVER_MODE_NORMAL = True


BITRATE_OBJS = {
    1000000: canstat.canBITRATE_1M,
    500000: canstat.canBITRATE_500K,
    250000: canstat.canBITRATE_250K,
    125000: canstat.canBITRATE_125K,
    100000: canstat.canBITRATE_100K,
    83000: canstat.canBITRATE_83K,
    62000: canstat.canBITRATE_62K,
    50000: canstat.canBITRATE_50K,
    10000: canstat.canBITRATE_10K,
}

BITRATE_FD = {
    500000: canstat.canFD_BITRATE_500K_80P,
    1000000: canstat.canFD_BITRATE_1M_80P,
    2000000: canstat.canFD_BITRATE_2M_80P,
    4000000: canstat.canFD_BITRATE_4M_80P,
    8000000: canstat.canFD_BITRATE_8M_60P,
}


class KvaserBus(BusABC):
    """
    The CAN Bus implemented for the Kvaser interface.
    """

    def __init__(self, channel, can_filters=None, **kwargs):
        """
        :param int channel:
            The Channel id to create this bus with.

        :param list can_filters:
            See :meth:`can.BusABC.set_filters`.

        Backend Configuration

        :param int bitrate:
            Bitrate of channel in bit/s
        :param bool accept_virtual:
            If virtual channels should be accepted.
        :param int tseg1:
            Time segment 1, that is, the number of quanta from (but not including)
            the Sync Segment to the sampling point.
            If this parameter is not given, the Kvaser driver will try to choose
            all bit timing parameters from a set of defaults.
        :param int tseg2:
            Time segment 2, that is, the number of quanta from the sampling
            point to the end of the bit.
        :param int sjw:
            The Synchronization Jump Width. Decides the maximum number of time quanta
            that the controller can resynchronize every bit.
        :param int no_samp:
            Either 1 or 3. Some CAN controllers can also sample each bit three times.
            In this case, the bit will be sampled three quanta in a row,
            with the last sample being taken in the edge between TSEG1 and TSEG2.
            Three samples should only be used for relatively slow baudrates.

        :param bool driver_mode:
            Silent or normal.

        :param bool single_handle:
            Use one Kvaser CANLIB bus handle for both reading and writing.
            This can be set if reading and/or writing is done from one thread.
        :param bool receive_own_messages:
            If messages transmitted should also be received back.
            Only works if single_handle is also False.
            If you want to receive messages from other applications on the same
            computer, set this to True or set single_handle to True.
        :param bool fd:
            If CAN-FD frames should be supported.
        :param int data_bitrate:
            Which bitrate to use for data phase in CAN FD.
            Defaults to arbitration bitrate.

        """

        log.info(f"CAN Filters: {can_filters}")
        log.info(f"Got configuration of: {kwargs}")
        bitrate = kwargs.get("bitrate", 500000)
        tseg1 = kwargs.get("tseg1", 0)
        tseg2 = kwargs.get("tseg2", 0)
        sjw = kwargs.get("sjw", 0)
        no_samp = kwargs.get("no_samp", 0)
        driver_mode = kwargs.get("driver_mode", DRIVER_MODE_NORMAL)
        single_handle = kwargs.get("single_handle", False)
        receive_own_messages = kwargs.get("receive_own_messages", False)
        accept_virtual = kwargs.get("accept_virtual", True)
        fd = kwargs.get("fd", False)
        data_bitrate = kwargs.get("data_bitrate", None)

        try:
            channel = int(channel)
        except ValueError:
            raise ValueError("channel must be an integer")
        self.channel = channel

        log.debug("Initialising bus instance")
        self.single_handle = single_handle

        num_channels = ctypes.c_int(0)
        canGetNumberOfChannels(ctypes.byref(num_channels))
        num_channels = int(num_channels.value)
        log.info("Found %d available channels", num_channels)
        for idx in range(num_channels):
            channel_info = get_channel_info(idx)
            log.info("%d: %s", idx, channel_info)
            if idx == channel:
                self.channel_info = channel_info

        flags = 0
        if accept_virtual:
            flags |= canstat.canOPEN_ACCEPT_VIRTUAL
        if fd:
            flags |= canstat.canOPEN_CAN_FD

        log.debug("Creating read handle to bus channel: %s", channel)
        self._read_handle = canOpenChannel(channel, flags)
        canIoCtlInit(
            self._read_handle,
            canstat.canIOCTL_SET_TIMER_SCALE,
            ctypes.byref(ctypes.c_long(TIMESTAMP_RESOLUTION)),
            4,
        )

        if fd:
            if "tseg1" not in kwargs and bitrate in BITRATE_FD:
                # Use predefined bitrate for arbitration
                bitrate = BITRATE_FD[bitrate]
            if data_bitrate in BITRATE_FD:
                # Use predefined bitrate for data
                data_bitrate = BITRATE_FD[data_bitrate]
            elif not data_bitrate:
                # Use same bitrate for arbitration and data phase
                data_bitrate = bitrate
            canSetBusParamsFd(self._read_handle, data_bitrate, tseg1, tseg2, sjw)
        else:
            if "tseg1" not in kwargs and bitrate in BITRATE_OBJS:
                bitrate = BITRATE_OBJS[bitrate]
        canSetBusParams(self._read_handle, bitrate, tseg1, tseg2, sjw, no_samp, 0)

        # By default, use local echo if single handle is used (see #160)
        local_echo = single_handle or receive_own_messages
        if receive_own_messages and single_handle:
            log.warning("receive_own_messages only works if single_handle is False")
        canIoCtlInit(
            self._read_handle,
            canstat.canIOCTL_SET_LOCAL_TXECHO,
            ctypes.byref(ctypes.c_byte(local_echo)),
            1,
        )

        if self.single_handle:
            log.debug("We don't require separate handles to the bus")
            self._write_handle = self._read_handle
        else:
            log.debug("Creating separate handle for TX on channel: %s", channel)
            self._write_handle = canOpenChannel(channel, flags)
            canBusOn(self._read_handle)

        can_driver_mode = (
            canstat.canDRIVER_SILENT
            if driver_mode == DRIVER_MODE_SILENT
            else canstat.canDRIVER_NORMAL
        )
        canSetBusOutputControl(self._write_handle, can_driver_mode)
        log.debug("Going bus on TX handle")
        canBusOn(self._write_handle)

        timer = ctypes.c_uint(0)
        try:
            if time.get_clock_info("time").resolution > 1e-5:
                ts, perfcounter = time_perfcounter_correlation()
                kvReadTimer(self._read_handle, ctypes.byref(timer))
                current_perfcounter = time.perf_counter()
                now = ts + (current_perfcounter - perfcounter)
                self._timestamp_offset = now - (timer.value * TIMESTAMP_FACTOR)
            else:
                kvReadTimer(self._read_handle, ctypes.byref(timer))
                self._timestamp_offset = time.time() - (timer.value * TIMESTAMP_FACTOR)

        except Exception as exc:
            # timer is usually close to 0
            log.info(str(exc))
            self._timestamp_offset = time.time() - (timer.value * TIMESTAMP_FACTOR)

        self._is_filtered = False
        super().__init__(channel=channel, can_filters=can_filters, **kwargs)

    def _apply_filters(self, filters):
        if filters and len(filters) == 1:
            can_id = filters[0]["can_id"]
            can_mask = filters[0]["can_mask"]
            extended = 1 if filters[0].get("extended") else 0
            try:
                for handle in (self._read_handle, self._write_handle):
                    canSetAcceptanceFilter(handle, can_id, can_mask, extended)
            except (NotImplementedError, CANLIBError) as e:
                self._is_filtered = False
                log.error("Filtering is not supported - %s", e)
            else:
                self._is_filtered = True
                log.info("canlib is filtering on ID 0x%X, mask 0x%X", can_id, can_mask)

        else:
            self._is_filtered = False
            log.info("Hardware filtering has been disabled")
            try:
                for handle in (self._read_handle, self._write_handle):
                    for extended in (0, 1):
                        canSetAcceptanceFilter(handle, 0, 0, extended)
            except (NotImplementedError, CANLIBError) as e:
                log.error("An error occured while disabling filtering: %s", e)

    def flush_tx_buffer(self):
        """Wipeout the transmit buffer on the Kvaser."""
        canIoCtl(self._write_handle, canstat.canIOCTL_FLUSH_TX_BUFFER, 0, 0)

    def _recv_internal(self, timeout=None):
        """
        Read a message from kvaser device and return whether filtering has taken place.
        """
        arb_id = ctypes.c_long(0)
        data = ctypes.create_string_buffer(64)
        dlc = ctypes.c_uint(0)
        flags = ctypes.c_uint(0)
        timestamp = ctypes.c_ulong(0)

        if timeout is None:
            # Set infinite timeout
            # http://www.kvaser.com/canlib-webhelp/group___c_a_n.html#ga2edd785a87cc16b49ece8969cad71e5b
            timeout = 0xFFFFFFFF
        else:
            timeout = int(timeout * 1000)

        # log.log(9, 'Reading for %d ms on handle: %s' % (timeout, self._read_handle))
        status = canReadWait(
            self._read_handle,
            ctypes.byref(arb_id),
            ctypes.byref(data),
            ctypes.byref(dlc),
            ctypes.byref(flags),
            ctypes.byref(timestamp),
            timeout,  # This is an X ms blocking read
        )

        if status == canstat.canOK:
            data_array = data.raw
            flags = flags.value
            is_extended = bool(flags & canstat.canMSG_EXT)
            is_remote_frame = bool(flags & canstat.canMSG_RTR)
            is_error_frame = bool(flags & canstat.canMSG_ERROR_FRAME)
            is_fd = bool(flags & canstat.canFDMSG_FDF)
            bitrate_switch = bool(flags & canstat.canFDMSG_BRS)
            error_state_indicator = bool(flags & canstat.canFDMSG_ESI)
            msg_timestamp = timestamp.value * TIMESTAMP_FACTOR
            rx_msg = Message(
                arbitration_id=arb_id.value,
                data=data_array[: dlc.value],
                dlc=dlc.value,
                is_extended_id=is_extended,
                is_error_frame=is_error_frame,
                is_remote_frame=is_remote_frame,
                is_fd=is_fd,
                bitrate_switch=bitrate_switch,
                error_state_indicator=error_state_indicator,
                channel=self.channel,
                timestamp=msg_timestamp + self._timestamp_offset,
            )
            # log.debug('Got message: %s' % rx_msg)
            return rx_msg, self._is_filtered
        else:
            # log.debug('read complete -> status not okay')
            return None, self._is_filtered

    def send(self, msg, timeout=None):
        # log.debug("Writing a message: {}".format(msg))
        flags = canstat.canMSG_EXT if msg.is_extended_id else canstat.canMSG_STD
        if msg.is_remote_frame:
            flags |= canstat.canMSG_RTR
        if msg.is_error_frame:
            flags |= canstat.canMSG_ERROR_FRAME
        if msg.is_fd:
            flags |= canstat.canFDMSG_FDF
        if msg.bitrate_switch:
            flags |= canstat.canFDMSG_BRS
        ArrayConstructor = ctypes.c_byte * msg.dlc
        buf = ArrayConstructor(*msg.data)
        canWrite(
            self._write_handle, msg.arbitration_id, ctypes.byref(buf), msg.dlc, flags
        )
        if timeout:
            canWriteSync(self._write_handle, int(timeout * 1000))

    def flash(self, flash=True):
        """
        Turn on or off flashing of the device's LED for physical
        identification purposes.
        """
        if flash:
            action = canstat.kvLED_ACTION_ALL_LEDS_ON
        else:
            action = canstat.kvLED_ACTION_ALL_LEDS_OFF

        try:
            kvFlashLeds(self._read_handle, action, 30000)
        except (CANLIBError, NotImplementedError) as e:
            log.error("Could not flash LEDs (%s)", e)

    def shutdown(self):
        super().shutdown()
        # Wait for transmit queue to be cleared
        try:
            canWriteSync(self._write_handle, 100)
        except CANLIBError:
            # Not a huge deal and it seems that we get timeout if no messages
            # exists in the buffer at all
            pass
        if not self.single_handle:
            canBusOff(self._read_handle)
            canClose(self._read_handle)
        canBusOff(self._write_handle)
        canClose(self._write_handle)

    def get_stats(self) -> structures.BusStatistics:
        """Retrieves the bus statistics.

        Use like so:

        .. testsetup:: kvaser

            from unittest.mock import Mock
            from can.interfaces.kvaser.structures import BusStatistics
            bus = Mock()
            bus.get_stats = Mock(side_effect=lambda: BusStatistics())

        .. doctest:: kvaser

            >>> stats = bus.get_stats()
            >>> print(stats)
            std_data: 0, std_remote: 0, ext_data: 0, ext_remote: 0, err_frame: 0, bus_load: 0.0%, overruns: 0

        :returns: bus statistics.
        """
        canRequestBusStatistics(self._write_handle)
        stats = structures.BusStatistics()
        canGetBusStatistics(
            self._write_handle, ctypes.pointer(stats), ctypes.sizeof(stats)
        )
        return stats

    @staticmethod
    def _detect_available_configs():
        num_channels = ctypes.c_int(0)
        try:
            canGetNumberOfChannels(ctypes.byref(num_channels))
        except (CANLIBError, NameError):
            pass

        return [
            {"interface": "kvaser", "channel": channel}
            for channel in range(num_channels.value)
        ]


def get_channel_info(channel):
    name = ctypes.create_string_buffer(80)
    serial = ctypes.c_uint64()
    number = ctypes.c_uint()

    canGetChannelData(
        channel,
        canstat.canCHANNELDATA_DEVDESCR_ASCII,
        ctypes.byref(name),
        ctypes.sizeof(name),
    )
    canGetChannelData(
        channel,
        canstat.canCHANNELDATA_CARD_SERIAL_NO,
        ctypes.byref(serial),
        ctypes.sizeof(serial),
    )
    canGetChannelData(
        channel,
        canstat.canCHANNELDATA_CHAN_NO_ON_CARD,
        ctypes.byref(number),
        ctypes.sizeof(number),
    )

    return f"{name.value.decode('ascii')}, S/N {serial.value} (#{number.value + 1})"


init_kvaser_library()



"""
Interface to socketcand
see https://github.com/linux-can/socketcand

Authors: Marvin Seiler, Gerrit Telkamp

Copyright (C) 2021  DOMOLOGIC GmbH
http://www.domologic.de
"""
import logging
import select
import socket
import time
import traceback
from collections import deque

import can

log = logging.getLogger(__name__)


def convert_ascii_message_to_can_message(ascii_msg: str) -> can.Message:
    if not ascii_msg.startswith("< frame ") or not ascii_msg.endswith(" >"):
        log.warning(f"Could not parse ascii message: {ascii_msg}")
        return None
    else:
        # frame_string = ascii_msg.removeprefix("< frame ").removesuffix(" >")
        frame_string = ascii_msg[8:-2]
        parts = frame_string.split(" ", 3)
        can_id, timestamp = int(parts[0], 16), float(parts[1])
        is_ext = len(parts[0]) != 3

        data = bytearray.fromhex(parts[2])
        can_dlc = len(data)
        can_message = can.Message(
            timestamp=timestamp,
            arbitration_id=can_id,
            data=data,
            dlc=can_dlc,
            is_extended_id=is_ext,
            is_rx=True,
        )
        return can_message


def convert_can_message_to_ascii_message(can_message: can.Message) -> str:
    # Note: socketcan bus adds extended flag, remote_frame_flag & error_flag to id
    # not sure if that is necessary here
    can_id = can_message.arbitration_id
    if can_message.is_extended_id:
        can_id_string = f"{(can_id&0x1FFFFFFF):08X}"
    else:
        can_id_string = f"{(can_id&0x7FF):03X}"
    # Note: seems like we cannot add CANFD_BRS (bitrate_switch) and CANFD_ESI (error_state_indicator) flags
    data = can_message.data
    length = can_message.dlc
    bytes_string = " ".join(f"{x:x}" for x in data[0:length])
    return f"< send {can_id_string} {length:X} {bytes_string} >"


def connect_to_server(s, host, port):
    timeout_ms = 10000
    now = time.time() * 1000
    end_time = now + timeout_ms
    while now < end_time:
        try:
            s.connect((host, port))
            return
        except Exception as e:
            log.warning(f"Failed to connect to server: {type(e)} Message: {e}")
            now = time.time() * 1000
    raise TimeoutError(
        f"connect_to_server: Failed to connect server for {timeout_ms} ms"
    )


class SocketCanDaemonBus(can.BusABC):
    def __init__(self, channel, host, port, can_filters=None, **kwargs):
        self.__host = host
        self.__port = port
        self.__socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.__message_buffer = deque()
        self.__receive_buffer = ""  # i know string is not the most efficient here
        self.channel = channel
        self.channel_info = f"socketcand on {channel}@{host}:{port}"
        connect_to_server(self.__socket, self.__host, self.__port)
        self._expect_msg("< hi >")

        log.info(
            f"SocketCanDaemonBus: connected with address {self.__socket.getsockname()}"
        )
        self._tcp_send(f"< open {channel} >")
        self._expect_msg("< ok >")
        self._tcp_send(f"< rawmode >")
        self._expect_msg("< ok >")
        super().__init__(channel=channel, can_filters=can_filters)

    def _recv_internal(self, timeout):
        if len(self.__message_buffer) != 0:
            can_message = self.__message_buffer.popleft()
            return can_message, False

        try:
            # get all sockets that are ready (can be a list with a single value
            # being self.socket or an empty list if self.socket is not ready)
            ready_receive_sockets, _, _ = select.select(
                [self.__socket], [], [], timeout
            )
        except OSError as exc:
            # something bad happened (e.g. the interface went down)
            log.error(f"Failed to receive: {exc}")
            raise can.CanError(f"Failed to receive: {exc}")

        try:
            if not ready_receive_sockets:
                # socket wasn't readable or timeout occurred
                log.debug("Socket not ready")
                return None, False

            ascii_msg = self.__socket.recv(1024).decode(
                "ascii"
            )  # may contain multiple messages
            self.__receive_buffer += ascii_msg
            log.debug(f"Received Ascii Message: {ascii_msg}")
            buffer_view = self.__receive_buffer
            chars_processed_successfully = 0
            while True:
                if len(buffer_view) == 0:
                    break

                start = buffer_view.find("<")
                if start == -1:
                    log.warning(
                        f"Bad data: No opening < found => discarding entire buffer '{buffer_view}'"
                    )
                    chars_processed_successfully = len(self.__receive_buffer)
                    break
                end = buffer_view.find(">")
                if end == -1:
                    log.warning("Got incomplete message => waiting for more data")
                    if len(buffer_view) > 200:
                        log.warning(
                            "Incomplete message exceeds 200 chars => Discarding"
                        )
                        chars_processed_successfully = len(self.__receive_buffer)
                    break
                chars_processed_successfully += end + 1
                single_message = buffer_view[start : end + 1]
                parsed_can_message = convert_ascii_message_to_can_message(
                    single_message
                )
                if parsed_can_message is None:
                    log.warning(f"Invalid Frame: {single_message}")
                else:
                    parsed_can_message.channel = self.channel
                    self.__message_buffer.append(parsed_can_message)
                buffer_view = buffer_view[end + 1 :]

            self.__receive_buffer = self.__receive_buffer[
                chars_processed_successfully + 1 :
            ]
            can_message = (
                None
                if len(self.__message_buffer) == 0
                else self.__message_buffer.popleft()
            )
            return can_message, False

        except Exception as exc:
            log.error(f"Failed to receive: {exc}  {traceback.format_exc()}")
            raise can.CanError(f"Failed to receive: {exc}  {traceback.format_exc()}")

    def _tcp_send(self, msg: str):
        log.debug(f"Sending TCP Message: '{msg}'")
        self.__socket.sendall(msg.encode("ascii"))

    def _expect_msg(self, msg):
        ascii_msg = self.__socket.recv(256).decode("ascii")
        if not ascii_msg == msg:
            raise can.CanError(f"{msg} message expected!")

    def send(self, msg, timeout=None):
        ascii_msg = convert_can_message_to_ascii_message(msg)
        self._tcp_send(ascii_msg)

    def shutdown(self):
        super().shutdown()
        self.__socket.close()



"""
"""

import logging
from typing import List

log = logging.getLogger("can.usb2can")

try:
    import win32com.client
except ImportError:
    log.warning("win32com.client module required for usb2can")
    raise


def WMIDateStringToDate(dtmDate) -> str:
    if dtmDate[4] == 0:
        strDateTime = dtmDate[5] + "/"
    else:
        strDateTime = dtmDate[4] + dtmDate[5] + "/"

    if dtmDate[6] == 0:
        strDateTime = strDateTime + dtmDate[7] + "/"
    else:
        strDateTime = strDateTime + dtmDate[6] + dtmDate[7] + "/"
        strDateTime = (
            strDateTime
            + dtmDate[0]
            + dtmDate[1]
            + dtmDate[2]
            + dtmDate[3]
            + " "
            + dtmDate[8]
            + dtmDate[9]
            + ":"
            + dtmDate[10]
            + dtmDate[11]
            + ":"
            + dtmDate[12]
            + dtmDate[13]
        )
    return strDateTime


def find_serial_devices(serial_matcher: str = "") -> List[str]:
    """
    Finds a list of USB devices where the serial number (partially) matches the given string.

    :param serial_matcher:
        only device IDs starting with this string are returned
    """
    serial_numbers = []
    wmi = win32com.client.GetObject("winmgmts:")
    for usb_controller in wmi.InstancesOf("Win32_USBControllerDevice"):
        usb_device = wmi.Get(usb_controller.Dependent)
        if "USB2CAN" in usb_device.Name:
            serial_numbers.append(usb_device.DeviceID.split("\\")[-1])

    if serial_matcher:
        return [sn for sn in serial_numbers if serial_matcher in sn]
    return serial_numbers



#  PCANBasic.py
#
#  ~~~~~~~~~~~~
#
#  PCAN-Basic API
#
#  ~~~~~~~~~~~~
#
#  ------------------------------------------------------------------
#  Author : Keneth Wagner
#  Last change: 2022-07-06
#  ------------------------------------------------------------------
#
#  Copyright (C) 1999-2022  PEAK-System Technik GmbH, Darmstadt
#  more Info at http://www.peak-system.com

# Module Imports
import logging
import platform
from ctypes import *
from ctypes.util import find_library

PLATFORM = platform.system()
IS_WINDOWS = PLATFORM == "Windows"
IS_LINUX = PLATFORM == "Linux"

if IS_WINDOWS:
    import winreg

logger = logging.getLogger("can.pcan")

# ///////////////////////////////////////////////////////////
# Type definitions
# ///////////////////////////////////////////////////////////

TPCANHandle = c_ushort  # Represents a PCAN hardware channel handle
TPCANStatus = int  # Represents a PCAN status/error code
TPCANParameter = c_ubyte  # Represents a PCAN parameter to be read or set
TPCANDevice = c_ubyte  # Represents a PCAN device
TPCANMessageType = c_ubyte  # Represents the type of a PCAN message
TPCANType = c_ubyte  # Represents the type of PCAN hardware to be initialized
TPCANMode = c_ubyte  # Represents a PCAN filter mode
TPCANBaudrate = c_ushort  # Represents a PCAN Baud rate register value
TPCANBitrateFD = c_char_p  # Represents a PCAN-FD bit rate string
TPCANTimestampFD = c_ulonglong  # Represents a timestamp of a received PCAN FD message

# ///////////////////////////////////////////////////////////
# Value definitions
# ///////////////////////////////////////////////////////////

# Currently defined and supported PCAN channels
#
PCAN_NONEBUS = TPCANHandle(0x00)  # Undefined/default value for a PCAN bus

PCAN_ISABUS1 = TPCANHandle(0x21)  # PCAN-ISA interface, channel 1
PCAN_ISABUS2 = TPCANHandle(0x22)  # PCAN-ISA interface, channel 2
PCAN_ISABUS3 = TPCANHandle(0x23)  # PCAN-ISA interface, channel 3
PCAN_ISABUS4 = TPCANHandle(0x24)  # PCAN-ISA interface, channel 4
PCAN_ISABUS5 = TPCANHandle(0x25)  # PCAN-ISA interface, channel 5
PCAN_ISABUS6 = TPCANHandle(0x26)  # PCAN-ISA interface, channel 6
PCAN_ISABUS7 = TPCANHandle(0x27)  # PCAN-ISA interface, channel 7
PCAN_ISABUS8 = TPCANHandle(0x28)  # PCAN-ISA interface, channel 8

PCAN_DNGBUS1 = TPCANHandle(0x31)  # PCAN-Dongle/LPT interface, channel 1

PCAN_PCIBUS1 = TPCANHandle(0x41)  # PCAN-PCI interface, channel 1
PCAN_PCIBUS2 = TPCANHandle(0x42)  # PCAN-PCI interface, channel 2
PCAN_PCIBUS3 = TPCANHandle(0x43)  # PCAN-PCI interface, channel 3
PCAN_PCIBUS4 = TPCANHandle(0x44)  # PCAN-PCI interface, channel 4
PCAN_PCIBUS5 = TPCANHandle(0x45)  # PCAN-PCI interface, channel 5
PCAN_PCIBUS6 = TPCANHandle(0x46)  # PCAN-PCI interface, channel 6
PCAN_PCIBUS7 = TPCANHandle(0x47)  # PCAN-PCI interface, channel 7
PCAN_PCIBUS8 = TPCANHandle(0x48)  # PCAN-PCI interface, channel 8
PCAN_PCIBUS9 = TPCANHandle(0x409)  # PCAN-PCI interface, channel 9
PCAN_PCIBUS10 = TPCANHandle(0x40A)  # PCAN-PCI interface, channel 10
PCAN_PCIBUS11 = TPCANHandle(0x40B)  # PCAN-PCI interface, channel 11
PCAN_PCIBUS12 = TPCANHandle(0x40C)  # PCAN-PCI interface, channel 12
PCAN_PCIBUS13 = TPCANHandle(0x40D)  # PCAN-PCI interface, channel 13
PCAN_PCIBUS14 = TPCANHandle(0x40E)  # PCAN-PCI interface, channel 14
PCAN_PCIBUS15 = TPCANHandle(0x40F)  # PCAN-PCI interface, channel 15
PCAN_PCIBUS16 = TPCANHandle(0x410)  # PCAN-PCI interface, channel 16

PCAN_USBBUS1 = TPCANHandle(0x51)  # PCAN-USB interface, channel 1
PCAN_USBBUS2 = TPCANHandle(0x52)  # PCAN-USB interface, channel 2
PCAN_USBBUS3 = TPCANHandle(0x53)  # PCAN-USB interface, channel 3
PCAN_USBBUS4 = TPCANHandle(0x54)  # PCAN-USB interface, channel 4
PCAN_USBBUS5 = TPCANHandle(0x55)  # PCAN-USB interface, channel 5
PCAN_USBBUS6 = TPCANHandle(0x56)  # PCAN-USB interface, channel 6
PCAN_USBBUS7 = TPCANHandle(0x57)  # PCAN-USB interface, channel 7
PCAN_USBBUS8 = TPCANHandle(0x58)  # PCAN-USB interface, channel 8
PCAN_USBBUS9 = TPCANHandle(0x509)  # PCAN-USB interface, channel 9
PCAN_USBBUS10 = TPCANHandle(0x50A)  # PCAN-USB interface, channel 10
PCAN_USBBUS11 = TPCANHandle(0x50B)  # PCAN-USB interface, channel 11
PCAN_USBBUS12 = TPCANHandle(0x50C)  # PCAN-USB interface, channel 12
PCAN_USBBUS13 = TPCANHandle(0x50D)  # PCAN-USB interface, channel 13
PCAN_USBBUS14 = TPCANHandle(0x50E)  # PCAN-USB interface, channel 14
PCAN_USBBUS15 = TPCANHandle(0x50F)  # PCAN-USB interface, channel 15
PCAN_USBBUS16 = TPCANHandle(0x510)  # PCAN-USB interface, channel 16

PCAN_PCCBUS1 = TPCANHandle(0x61)  # PCAN-PC Card interface, channel 1
PCAN_PCCBUS2 = TPCANHandle(0x62)  # PCAN-PC Card interface, channel 2

PCAN_LANBUS1 = TPCANHandle(0x801)  # PCAN-LAN interface, channel 1
PCAN_LANBUS2 = TPCANHandle(0x802)  # PCAN-LAN interface, channel 2
PCAN_LANBUS3 = TPCANHandle(0x803)  # PCAN-LAN interface, channel 3
PCAN_LANBUS4 = TPCANHandle(0x804)  # PCAN-LAN interface, channel 4
PCAN_LANBUS5 = TPCANHandle(0x805)  # PCAN-LAN interface, channel 5
PCAN_LANBUS6 = TPCANHandle(0x806)  # PCAN-LAN interface, channel 6
PCAN_LANBUS7 = TPCANHandle(0x807)  # PCAN-LAN interface, channel 7
PCAN_LANBUS8 = TPCANHandle(0x808)  # PCAN-LAN interface, channel 8
PCAN_LANBUS9 = TPCANHandle(0x809)  # PCAN-LAN interface, channel 9
PCAN_LANBUS10 = TPCANHandle(0x80A)  # PCAN-LAN interface, channel 10
PCAN_LANBUS11 = TPCANHandle(0x80B)  # PCAN-LAN interface, channel 11
PCAN_LANBUS12 = TPCANHandle(0x80C)  # PCAN-LAN interface, channel 12
PCAN_LANBUS13 = TPCANHandle(0x80D)  # PCAN-LAN interface, channel 13
PCAN_LANBUS14 = TPCANHandle(0x80E)  # PCAN-LAN interface, channel 14
PCAN_LANBUS15 = TPCANHandle(0x80F)  # PCAN-LAN interface, channel 15
PCAN_LANBUS16 = TPCANHandle(0x810)  # PCAN-LAN interface, channel 16

# Represent the PCAN error and status codes
#
PCAN_ERROR_OK = TPCANStatus(0x00000)  # No error
PCAN_ERROR_XMTFULL = TPCANStatus(0x00001)  # Transmit buffer in CAN controller is full
PCAN_ERROR_OVERRUN = TPCANStatus(0x00002)  # CAN controller was read too late
PCAN_ERROR_BUSLIGHT = TPCANStatus(
    0x00004
)  # Bus error: an error counter reached the 'light' limit
PCAN_ERROR_BUSHEAVY = TPCANStatus(
    0x00008
)  # Bus error: an error counter reached the 'heavy' limit
PCAN_ERROR_BUSWARNING = TPCANStatus(
    PCAN_ERROR_BUSHEAVY
)  # Bus error: an error counter reached the 'warning' limit
PCAN_ERROR_BUSPASSIVE = TPCANStatus(
    0x40000
)  # Bus error: the CAN controller is error passive
PCAN_ERROR_BUSOFF = TPCANStatus(
    0x00010
)  # Bus error: the CAN controller is in bus-off state
PCAN_ERROR_ANYBUSERR = TPCANStatus(
    PCAN_ERROR_BUSWARNING
    | PCAN_ERROR_BUSLIGHT
    | PCAN_ERROR_BUSHEAVY
    | PCAN_ERROR_BUSOFF
    | PCAN_ERROR_BUSPASSIVE
)  # Mask for all bus errors
PCAN_ERROR_QRCVEMPTY = TPCANStatus(0x00020)  # Receive queue is empty
PCAN_ERROR_QOVERRUN = TPCANStatus(0x00040)  # Receive queue was read too late
PCAN_ERROR_QXMTFULL = TPCANStatus(0x00080)  # Transmit queue is full
PCAN_ERROR_REGTEST = TPCANStatus(
    0x00100
)  # Test of the CAN controller hardware registers failed (no hardware found)
PCAN_ERROR_NODRIVER = TPCANStatus(0x00200)  # Driver not loaded
PCAN_ERROR_HWINUSE = TPCANStatus(0x00400)  # Hardware already in use by a Net
PCAN_ERROR_NETINUSE = TPCANStatus(0x00800)  # A Client is already connected to the Net
PCAN_ERROR_ILLHW = TPCANStatus(0x01400)  # Hardware handle is invalid
PCAN_ERROR_ILLNET = TPCANStatus(0x01800)  # Net handle is invalid
PCAN_ERROR_ILLCLIENT = TPCANStatus(0x01C00)  # Client handle is invalid
PCAN_ERROR_ILLHANDLE = TPCANStatus(
    PCAN_ERROR_ILLHW | PCAN_ERROR_ILLNET | PCAN_ERROR_ILLCLIENT
)  # Mask for all handle errors
PCAN_ERROR_RESOURCE = TPCANStatus(
    0x02000
)  # Resource (FIFO, Client, timeout) cannot be created
PCAN_ERROR_ILLPARAMTYPE = TPCANStatus(0x04000)  # Invalid parameter
PCAN_ERROR_ILLPARAMVAL = TPCANStatus(0x08000)  # Invalid parameter value
PCAN_ERROR_UNKNOWN = TPCANStatus(0x10000)  # Unknown error
PCAN_ERROR_ILLDATA = TPCANStatus(0x20000)  # Invalid data, function, or action
PCAN_ERROR_ILLMODE = TPCANStatus(
    0x80000
)  # Driver object state is wrong for the attempted operation
PCAN_ERROR_CAUTION = TPCANStatus(
    0x2000000
)  # An operation was successfully carried out, however, irregularities were registered
PCAN_ERROR_INITIALIZE = TPCANStatus(
    0x4000000
)  # Channel is not initialized [Value was changed from 0x40000 to 0x4000000]
PCAN_ERROR_ILLOPERATION = TPCANStatus(
    0x8000000
)  # Invalid operation [Value was changed from 0x80000 to 0x8000000]

# PCAN devices
#
PCAN_NONE = TPCANDevice(0x00)  # Undefined, unknown or not selected PCAN device value
PCAN_PEAKCAN = TPCANDevice(0x01)  # PCAN Non-PnP devices. NOT USED WITHIN PCAN-Basic API
PCAN_ISA = TPCANDevice(0x02)  # PCAN-ISA, PCAN-PC/104, and PCAN-PC/104-Plus
PCAN_DNG = TPCANDevice(0x03)  # PCAN-Dongle
PCAN_PCI = TPCANDevice(0x04)  # PCAN-PCI, PCAN-cPCI, PCAN-miniPCI, and PCAN-PCI Express
PCAN_USB = TPCANDevice(0x05)  # PCAN-USB and PCAN-USB Pro
PCAN_PCC = TPCANDevice(0x06)  # PCAN-PC Card
PCAN_VIRTUAL = TPCANDevice(
    0x07
)  # PCAN Virtual hardware. NOT USED WITHIN PCAN-Basic API
PCAN_LAN = TPCANDevice(0x08)  # PCAN Gateway devices

# PCAN parameters
#
PCAN_DEVICE_ID = TPCANParameter(0x01)  # Device identifier parameter
PCAN_5VOLTS_POWER = TPCANParameter(0x02)  # 5-Volt power parameter
PCAN_RECEIVE_EVENT = TPCANParameter(0x03)  # PCAN receive event handler parameter
PCAN_MESSAGE_FILTER = TPCANParameter(0x04)  # PCAN message filter parameter
PCAN_API_VERSION = TPCANParameter(0x05)  # PCAN-Basic API version parameter
PCAN_CHANNEL_VERSION = TPCANParameter(0x06)  # PCAN device channel version parameter
PCAN_BUSOFF_AUTORESET = TPCANParameter(0x07)  # PCAN Reset-On-Busoff parameter
PCAN_LISTEN_ONLY = TPCANParameter(0x08)  # PCAN Listen-Only parameter
PCAN_LOG_LOCATION = TPCANParameter(0x09)  # Directory path for log files
PCAN_LOG_STATUS = TPCANParameter(0x0A)  # Debug-Log activation status
PCAN_LOG_CONFIGURE = TPCANParameter(
    0x0B
)  # Configuration of the debugged information (LOG_FUNCTION_***)
PCAN_LOG_TEXT = TPCANParameter(0x0C)  # Custom insertion of text into the log file
PCAN_CHANNEL_CONDITION = TPCANParameter(0x0D)  # Availability status of a PCAN-Channel
PCAN_HARDWARE_NAME = TPCANParameter(0x0E)  # PCAN hardware name parameter
PCAN_RECEIVE_STATUS = TPCANParameter(0x0F)  # Message reception status of a PCAN-Channel
PCAN_CONTROLLER_NUMBER = TPCANParameter(0x10)  # CAN-Controller number of a PCAN-Channel
PCAN_TRACE_LOCATION = TPCANParameter(0x11)  # Directory path for PCAN trace files
PCAN_TRACE_STATUS = TPCANParameter(0x12)  # CAN tracing activation status
PCAN_TRACE_SIZE = TPCANParameter(
    0x13
)  # Configuration of the maximum file size of a CAN trace
PCAN_TRACE_CONFIGURE = TPCANParameter(
    0x14
)  # Configuration of the trace file storing mode (TRACE_FILE_***)
PCAN_CHANNEL_IDENTIFYING = TPCANParameter(
    0x15
)  # Physical identification of a USB based PCAN-Channel by blinking its associated LED
PCAN_CHANNEL_FEATURES = TPCANParameter(
    0x16
)  # Capabilities of a PCAN device (FEATURE_***)
PCAN_BITRATE_ADAPTING = TPCANParameter(
    0x17
)  # Using of an existing bit rate (PCAN-View connected to a channel)
PCAN_BITRATE_INFO = TPCANParameter(0x18)  # Configured bit rate as Btr0Btr1 value
PCAN_BITRATE_INFO_FD = TPCANParameter(
    0x19
)  # Configured bit rate as TPCANBitrateFD string
PCAN_BUSSPEED_NOMINAL = TPCANParameter(
    0x1A
)  # Configured nominal CAN Bus speed as Bits per seconds
PCAN_BUSSPEED_DATA = TPCANParameter(
    0x1B
)  # Configured CAN data speed as Bits per seconds
PCAN_IP_ADDRESS = TPCANParameter(
    0x1C
)  # Remote address of a LAN channel as string in IPv4 format
PCAN_LAN_SERVICE_STATUS = TPCANParameter(
    0x1D
)  # Status of the Virtual PCAN-Gateway Service
PCAN_ALLOW_STATUS_FRAMES = TPCANParameter(
    0x1E
)  # Status messages reception status within a PCAN-Channel
PCAN_ALLOW_RTR_FRAMES = TPCANParameter(
    0x1F
)  # RTR messages reception status within a PCAN-Channel
PCAN_ALLOW_ERROR_FRAMES = TPCANParameter(
    0x20
)  # Error messages reception status within a PCAN-Channel
PCAN_INTERFRAME_DELAY = TPCANParameter(
    0x21
)  # Delay, in microseconds, between sending frames
PCAN_ACCEPTANCE_FILTER_11BIT = TPCANParameter(
    0x22
)  # Filter over code and mask patterns for 11-Bit messages
PCAN_ACCEPTANCE_FILTER_29BIT = TPCANParameter(
    0x23
)  # Filter over code and mask patterns for 29-Bit messages
PCAN_IO_DIGITAL_CONFIGURATION = TPCANParameter(
    0x24
)  # Output mode of 32 digital I/O pin of a PCAN-USB Chip. 1: Output-Active 0 : Output Inactive
PCAN_IO_DIGITAL_VALUE = TPCANParameter(
    0x25
)  # Value assigned to a 32 digital I/O pins of a PCAN-USB Chip
PCAN_IO_DIGITAL_SET = TPCANParameter(
    0x26
)  # Value assigned to a 32 digital I/O pins of a PCAN-USB Chip - Multiple digital I/O pins to 1 = High
PCAN_IO_DIGITAL_CLEAR = TPCANParameter(0x27)  # Clear multiple digital I/O pins to 0
PCAN_IO_ANALOG_VALUE = TPCANParameter(0x28)  # Get value of a single analog input pin
PCAN_FIRMWARE_VERSION = TPCANParameter(
    0x29
)  # Get the version of the firmware used by the device associated with a PCAN-Channel
PCAN_ATTACHED_CHANNELS_COUNT = TPCANParameter(
    0x2A
)  # Get the amount of PCAN channels attached to a system
PCAN_ATTACHED_CHANNELS = TPCANParameter(
    0x2B
)  # Get information about PCAN channels attached to a system
PCAN_ALLOW_ECHO_FRAMES = TPCANParameter(
    0x2C
)  # Echo messages reception status within a PCAN-Channel
PCAN_DEVICE_PART_NUMBER = TPCANParameter(
    0x2D
)  # Get the part number associated to a device

# DEPRECATED parameters
#
PCAN_DEVICE_NUMBER = PCAN_DEVICE_ID  # DEPRECATED. Use PCAN_DEVICE_ID instead

# PCAN parameter values
#
PCAN_PARAMETER_OFF = int(0x00)  # The PCAN parameter is not set (inactive)
PCAN_PARAMETER_ON = int(0x01)  # The PCAN parameter is set (active)
PCAN_FILTER_CLOSE = int(0x00)  # The PCAN filter is closed. No messages will be received
PCAN_FILTER_OPEN = int(
    0x01
)  # The PCAN filter is fully opened. All messages will be received
PCAN_FILTER_CUSTOM = int(
    0x02
)  # The PCAN filter is custom configured. Only registered messages will be received
PCAN_CHANNEL_UNAVAILABLE = int(
    0x00
)  # The PCAN-Channel handle is illegal, or its associated hardware is not available
PCAN_CHANNEL_AVAILABLE = int(
    0x01
)  # The PCAN-Channel handle is available to be connected (PnP Hardware: it means furthermore that the hardware is plugged-in)
PCAN_CHANNEL_OCCUPIED = int(
    0x02
)  # The PCAN-Channel handle is valid, and is already being used
PCAN_CHANNEL_PCANVIEW = (
    PCAN_CHANNEL_AVAILABLE | PCAN_CHANNEL_OCCUPIED
)  # The PCAN-Channel handle is already being used by a PCAN-View application, but is available to connect

LOG_FUNCTION_DEFAULT = int(0x00)  # Logs system exceptions / errors
LOG_FUNCTION_ENTRY = int(0x01)  # Logs the entries to the PCAN-Basic API functions
LOG_FUNCTION_PARAMETERS = int(
    0x02
)  # Logs the parameters passed to the PCAN-Basic API functions
LOG_FUNCTION_LEAVE = int(0x04)  # Logs the exits from the PCAN-Basic API functions
LOG_FUNCTION_WRITE = int(0x08)  # Logs the CAN messages passed to the CAN_Write function
LOG_FUNCTION_READ = int(
    0x10
)  # Logs the CAN messages received within the CAN_Read function
LOG_FUNCTION_ALL = int(
    0xFFFF
)  # Logs all possible information within the PCAN-Basic API functions

TRACE_FILE_SINGLE = int(
    0x00
)  # A single file is written until it size reaches PAN_TRACE_SIZE
TRACE_FILE_SEGMENTED = int(
    0x01
)  # Traced data is distributed in several files with size PAN_TRACE_SIZE
TRACE_FILE_DATE = int(0x02)  # Includes the date into the name of the trace file
TRACE_FILE_TIME = int(0x04)  # Includes the start time into the name of the trace file
TRACE_FILE_OVERWRITE = int(
    0x80
)  # Causes the overwriting of available traces (same name)

FEATURE_FD_CAPABLE = int(0x01)  # Device supports flexible data-rate (CAN-FD)
FEATURE_DELAY_CAPABLE = int(
    0x02
)  # Device supports a delay between sending frames (FPGA based USB devices)
FEATURE_IO_CAPABLE = int(
    0x04
)  # Device supports I/O functionality for electronic circuits (USB-Chip devices)

SERVICE_STATUS_STOPPED = int(0x01)  # The service is not running
SERVICE_STATUS_RUNNING = int(0x04)  # The service is running

# Other constants
#
MAX_LENGTH_HARDWARE_NAME = int(
    33
)  # Maximum length of the name of a device: 32 characters + terminator
MAX_LENGTH_VERSION_STRING = int(
    256
)  # Maximum length of a version string: 255 characters + terminator

# PCAN message types
#
PCAN_MESSAGE_STANDARD = TPCANMessageType(
    0x00
)  # The PCAN message is a CAN Standard Frame (11-bit identifier)
PCAN_MESSAGE_RTR = TPCANMessageType(
    0x01
)  # The PCAN message is a CAN Remote-Transfer-Request Frame
PCAN_MESSAGE_EXTENDED = TPCANMessageType(
    0x02
)  # The PCAN message is a CAN Extended Frame (29-bit identifier)
PCAN_MESSAGE_FD = TPCANMessageType(
    0x04
)  # The PCAN message represents a FD frame in terms of CiA Specs
PCAN_MESSAGE_BRS = TPCANMessageType(
    0x08
)  # The PCAN message represents a FD bit rate switch (CAN data at a higher bit rate)
PCAN_MESSAGE_ESI = TPCANMessageType(
    0x10
)  # The PCAN message represents a FD error state indicator(CAN FD transmitter was error active)
PCAN_MESSAGE_ECHO = TPCANMessageType(
    0x20
)  # The PCAN message represents an echo CAN Frame
PCAN_MESSAGE_ERRFRAME = TPCANMessageType(
    0x40
)  # The PCAN message represents an error frame
PCAN_MESSAGE_STATUS = TPCANMessageType(
    0x80
)  # The PCAN message represents a PCAN status message

# LookUp Parameters
#
LOOKUP_DEVICE_TYPE = (
    b"devicetype"  # Lookup channel by Device type (see PCAN devices e.g. PCAN_USB)
)
LOOKUP_DEVICE_ID = b"deviceid"  # Lookup channel by device id
LOOKUP_CONTROLLER_NUMBER = (
    b"controllernumber"  # Lookup channel by CAN controller 0-based index
)
LOOKUP_IP_ADDRESS = b"ipaddress"  # Lookup channel by IP address (LAN channels only)

# Frame Type / Initialization Mode
#
PCAN_MODE_STANDARD = PCAN_MESSAGE_STANDARD
PCAN_MODE_EXTENDED = PCAN_MESSAGE_EXTENDED

# Baud rate codes = BTR0/BTR1 register values for the CAN controller.
# You can define your own Baud rate with the BTROBTR1 register.
# Take a look at www.peak-system.com for our free software "BAUDTOOL"
# to calculate the BTROBTR1 register for every bit rate and sample point.
#
PCAN_BAUD_1M = TPCANBaudrate(0x0014)  #   1 MBit/s
PCAN_BAUD_800K = TPCANBaudrate(0x0016)  # 800 kBit/s
PCAN_BAUD_500K = TPCANBaudrate(0x001C)  # 500 kBit/s
PCAN_BAUD_250K = TPCANBaudrate(0x011C)  # 250 kBit/s
PCAN_BAUD_125K = TPCANBaudrate(0x031C)  # 125 kBit/s
PCAN_BAUD_100K = TPCANBaudrate(0x432F)  # 100 kBit/s
PCAN_BAUD_95K = TPCANBaudrate(0xC34E)  #  95,238 kBit/s
PCAN_BAUD_83K = TPCANBaudrate(0x852B)  #  83,333 kBit/s
PCAN_BAUD_50K = TPCANBaudrate(0x472F)  #  50 kBit/s
PCAN_BAUD_47K = TPCANBaudrate(0x1414)  #  47,619 kBit/s
PCAN_BAUD_33K = TPCANBaudrate(0x8B2F)  #  33,333 kBit/s
PCAN_BAUD_20K = TPCANBaudrate(0x532F)  #  20 kBit/s
PCAN_BAUD_10K = TPCANBaudrate(0x672F)  #  10 kBit/s
PCAN_BAUD_5K = TPCANBaudrate(0x7F7F)  #   5 kBit/s

# Represents the configuration for a CAN bit rate
# Note:
#    * Each parameter and its value must be separated with a '='.
#    * Each pair of parameter/value must be separated using ','.
#
# Example:
#    f_clock=80000000,nom_brp=10,nom_tseg1=5,nom_tseg2=2,nom_sjw=1,data_brp=4,data_tseg1=7,data_tseg2=2,data_sjw=1
#
PCAN_BR_CLOCK = TPCANBitrateFD(b"f_clock")
PCAN_BR_CLOCK_MHZ = TPCANBitrateFD(b"f_clock_mhz")
PCAN_BR_NOM_BRP = TPCANBitrateFD(b"nom_brp")
PCAN_BR_NOM_TSEG1 = TPCANBitrateFD(b"nom_tseg1")
PCAN_BR_NOM_TSEG2 = TPCANBitrateFD(b"nom_tseg2")
PCAN_BR_NOM_SJW = TPCANBitrateFD(b"nom_sjw")
PCAN_BR_NOM_SAMPLE = TPCANBitrateFD(b"nom_sam")
PCAN_BR_DATA_BRP = TPCANBitrateFD(b"data_brp")
PCAN_BR_DATA_TSEG1 = TPCANBitrateFD(b"data_tseg1")
PCAN_BR_DATA_TSEG2 = TPCANBitrateFD(b"data_tseg2")
PCAN_BR_DATA_SJW = TPCANBitrateFD(b"data_sjw")
PCAN_BR_DATA_SAMPLE = TPCANBitrateFD(b"data_ssp_offset")

# Supported Non-PnP Hardware types
#
PCAN_TYPE_ISA = TPCANType(0x01)  # PCAN-ISA 82C200
PCAN_TYPE_ISA_SJA = TPCANType(0x09)  # PCAN-ISA SJA1000
PCAN_TYPE_ISA_PHYTEC = TPCANType(0x04)  # PHYTEC ISA
PCAN_TYPE_DNG = TPCANType(0x02)  # PCAN-Dongle 82C200
PCAN_TYPE_DNG_EPP = TPCANType(0x03)  # PCAN-Dongle EPP 82C200
PCAN_TYPE_DNG_SJA = TPCANType(0x05)  # PCAN-Dongle SJA1000
PCAN_TYPE_DNG_SJA_EPP = TPCANType(0x06)  # PCAN-Dongle EPP SJA1000

# string description of the error codes
PCAN_DICT_STATUS = {
    PCAN_ERROR_OK: "OK",
    PCAN_ERROR_XMTFULL: "XMTFULL",
    PCAN_ERROR_OVERRUN: "OVERRUN",
    PCAN_ERROR_BUSLIGHT: "BUSLIGHT",
    PCAN_ERROR_BUSHEAVY: "BUSHEAVY",
    PCAN_ERROR_BUSWARNING: "BUSWARNING",
    PCAN_ERROR_BUSPASSIVE: "BUSPASSIVE",
    PCAN_ERROR_BUSOFF: "BUSOFF",
    PCAN_ERROR_ANYBUSERR: "ANYBUSERR",
    PCAN_ERROR_QRCVEMPTY: "QRCVEMPTY",
    PCAN_ERROR_QOVERRUN: "QOVERRUN",
    PCAN_ERROR_QXMTFULL: "QXMTFULL",
    PCAN_ERROR_REGTEST: "ERR_REGTEST",
    PCAN_ERROR_NODRIVER: "NODRIVER",
    PCAN_ERROR_HWINUSE: "HWINUSE",
    PCAN_ERROR_NETINUSE: "NETINUSE",
    PCAN_ERROR_ILLHW: "ILLHW",
    PCAN_ERROR_ILLNET: "ILLNET",
    PCAN_ERROR_ILLCLIENT: "ILLCLIENT",
    PCAN_ERROR_ILLHANDLE: "ILLHANDLE",
    PCAN_ERROR_RESOURCE: "ERR_RESOURCE",
    PCAN_ERROR_ILLPARAMTYPE: "ILLPARAMTYPE",
    PCAN_ERROR_ILLPARAMVAL: "ILLPARAMVAL",
    PCAN_ERROR_UNKNOWN: "UNKNOWN",
    PCAN_ERROR_ILLDATA: "ILLDATA",
    PCAN_ERROR_CAUTION: "CAUTION",
    PCAN_ERROR_INITIALIZE: "ERR_INITIALIZE",
    PCAN_ERROR_ILLOPERATION: "ILLOPERATION",
}


# Represents a PCAN message
#
class TPCANMsg(Structure):
    """
    Represents a PCAN message
    """

    _fields_ = [
        ("ID", c_uint),  # 11/29-bit message identifier
        ("MSGTYPE", TPCANMessageType),  # Type of the message
        ("LEN", c_ubyte),  # Data Length Code of the message (0..8)
        ("DATA", c_ubyte * 8),
    ]  # Data of the message (DATA[0]..DATA[7])


# Represents a timestamp of a received PCAN message
# Total Microseconds = micros + 1000 * millis + 0x100000000 * 1000 * millis_overflow
#
class TPCANTimestamp(Structure):
    """
    Represents a timestamp of a received PCAN message
    Total Microseconds = micros + 1000 * millis + 0x100000000 * 1000 * millis_overflow
    """

    _fields_ = [
        ("millis", c_uint),  # Base-value: milliseconds: 0.. 2^32-1
        ("millis_overflow", c_ushort),  # Roll-arounds of millis
        ("micros", c_ushort),
    ]  # Microseconds: 0..999


# Represents a PCAN message from a FD capable hardware
#
class TPCANMsgFD(Structure):
    """
    Represents a PCAN message
    """

    _fields_ = [
        ("ID", c_uint),  # 11/29-bit message identifier
        ("MSGTYPE", TPCANMessageType),  # Type of the message
        ("DLC", c_ubyte),  # Data Length Code of the message (0..15)
        ("DATA", c_ubyte * 64),
    ]  # Data of the message (DATA[0]..DATA[63])


# Describes an available PCAN channel
#
class TPCANChannelInformation(Structure):
    """
    Describes an available PCAN channel
    """

    _fields_ = [
        ("channel_handle", TPCANHandle),  # PCAN channel handle
        ("device_type", TPCANDevice),  # Kind of PCAN device
        ("controller_number", c_ubyte),  # CAN-Controller number
        ("device_features", c_uint),  # Device capabilities flag (see FEATURE_*)
        ("device_name", c_char * MAX_LENGTH_HARDWARE_NAME),  # Device name
        ("device_id", c_uint),  # Device number
        ("channel_condition", c_uint),
    ]  # Availability status of a PCAN-Channel


# ///////////////////////////////////////////////////////////
# Additional objects
# ///////////////////////////////////////////////////////////

PCAN_BITRATES = {
    1000000: PCAN_BAUD_1M,
    800000: PCAN_BAUD_800K,
    500000: PCAN_BAUD_500K,
    250000: PCAN_BAUD_250K,
    125000: PCAN_BAUD_125K,
    100000: PCAN_BAUD_100K,
    95000: PCAN_BAUD_95K,
    83000: PCAN_BAUD_83K,
    50000: PCAN_BAUD_50K,
    47000: PCAN_BAUD_47K,
    33000: PCAN_BAUD_33K,
    20000: PCAN_BAUD_20K,
    10000: PCAN_BAUD_10K,
    5000: PCAN_BAUD_5K,
}

PCAN_FD_PARAMETER_LIST = (
    "nom_brp",
    "nom_tseg1",
    "nom_tseg2",
    "nom_sjw",
    "data_brp",
    "data_tseg1",
    "data_tseg2",
    "data_sjw",
)

PCAN_CHANNEL_NAMES = {
    "PCAN_NONEBUS": PCAN_NONEBUS,
    "PCAN_ISABUS1": PCAN_ISABUS1,
    "PCAN_ISABUS2": PCAN_ISABUS2,
    "PCAN_ISABUS3": PCAN_ISABUS3,
    "PCAN_ISABUS4": PCAN_ISABUS4,
    "PCAN_ISABUS5": PCAN_ISABUS5,
    "PCAN_ISABUS6": PCAN_ISABUS6,
    "PCAN_ISABUS7": PCAN_ISABUS7,
    "PCAN_ISABUS8": PCAN_ISABUS8,
    "PCAN_DNGBUS1": PCAN_DNGBUS1,
    "PCAN_PCIBUS1": PCAN_PCIBUS1,
    "PCAN_PCIBUS2": PCAN_PCIBUS2,
    "PCAN_PCIBUS3": PCAN_PCIBUS3,
    "PCAN_PCIBUS4": PCAN_PCIBUS4,
    "PCAN_PCIBUS5": PCAN_PCIBUS5,
    "PCAN_PCIBUS6": PCAN_PCIBUS6,
    "PCAN_PCIBUS7": PCAN_PCIBUS7,
    "PCAN_PCIBUS8": PCAN_PCIBUS8,
    "PCAN_PCIBUS9": PCAN_PCIBUS9,
    "PCAN_PCIBUS10": PCAN_PCIBUS10,
    "PCAN_PCIBUS11": PCAN_PCIBUS11,
    "PCAN_PCIBUS12": PCAN_PCIBUS12,
    "PCAN_PCIBUS13": PCAN_PCIBUS13,
    "PCAN_PCIBUS14": PCAN_PCIBUS14,
    "PCAN_PCIBUS15": PCAN_PCIBUS15,
    "PCAN_PCIBUS16": PCAN_PCIBUS16,
    "PCAN_USBBUS1": PCAN_USBBUS1,
    "PCAN_USBBUS2": PCAN_USBBUS2,
    "PCAN_USBBUS3": PCAN_USBBUS3,
    "PCAN_USBBUS4": PCAN_USBBUS4,
    "PCAN_USBBUS5": PCAN_USBBUS5,
    "PCAN_USBBUS6": PCAN_USBBUS6,
    "PCAN_USBBUS7": PCAN_USBBUS7,
    "PCAN_USBBUS8": PCAN_USBBUS8,
    "PCAN_USBBUS9": PCAN_USBBUS9,
    "PCAN_USBBUS10": PCAN_USBBUS10,
    "PCAN_USBBUS11": PCAN_USBBUS11,
    "PCAN_USBBUS12": PCAN_USBBUS12,
    "PCAN_USBBUS13": PCAN_USBBUS13,
    "PCAN_USBBUS14": PCAN_USBBUS14,
    "PCAN_USBBUS15": PCAN_USBBUS15,
    "PCAN_USBBUS16": PCAN_USBBUS16,
    "PCAN_PCCBUS1": PCAN_PCCBUS1,
    "PCAN_PCCBUS2": PCAN_PCCBUS2,
    "PCAN_LANBUS1": PCAN_LANBUS1,
    "PCAN_LANBUS2": PCAN_LANBUS2,
    "PCAN_LANBUS3": PCAN_LANBUS3,
    "PCAN_LANBUS4": PCAN_LANBUS4,
    "PCAN_LANBUS5": PCAN_LANBUS5,
    "PCAN_LANBUS6": PCAN_LANBUS6,
    "PCAN_LANBUS7": PCAN_LANBUS7,
    "PCAN_LANBUS8": PCAN_LANBUS8,
    "PCAN_LANBUS9": PCAN_LANBUS9,
    "PCAN_LANBUS10": PCAN_LANBUS10,
    "PCAN_LANBUS11": PCAN_LANBUS11,
    "PCAN_LANBUS12": PCAN_LANBUS12,
    "PCAN_LANBUS13": PCAN_LANBUS13,
    "PCAN_LANBUS14": PCAN_LANBUS14,
    "PCAN_LANBUS15": PCAN_LANBUS15,
    "PCAN_LANBUS16": PCAN_LANBUS16,
}

VALID_PCAN_CAN_CLOCKS = [8_000_000]

VALID_PCAN_FD_CLOCKS = [
    20_000_000,
    24_000_000,
    30_000_000,
    40_000_000,
    60_000_000,
    80_000_000,
]

# ///////////////////////////////////////////////////////////
# PCAN-Basic API function declarations
# ///////////////////////////////////////////////////////////


# PCAN-Basic API class implementation
#
class PCANBasic:
    """PCAN-Basic API class implementation"""

    def __init__(self):
        if platform.system() == "Windows":
            load_library_func = windll.LoadLibrary

            # look for Peak drivers in Windows registry
            with winreg.ConnectRegistry(None, winreg.HKEY_LOCAL_MACHINE) as reg:
                try:
                    with winreg.OpenKey(reg, r"SOFTWARE\PEAK-System\PEAK-Drivers"):
                        pass
                except OSError:
                    raise OSError("The PEAK-driver could not be found!") from None
        else:
            load_library_func = cdll.LoadLibrary

        if platform.system() == "Windows" or "CYGWIN" in platform.system():
            lib_name = "PCANBasic.dll"
        elif platform.system() == "Darwin":
            # PCBUSB library is a third-party software created
            # and maintained by the MacCAN project
            lib_name = "libPCBUSB.dylib"
        else:
            lib_name = "libpcanbasic.so"

        lib_path = find_library(lib_name)
        if not lib_path:
            raise OSError(f"{lib_name} library not found.")

        try:
            self.__m_dllBasic = load_library_func(lib_path)
        except OSError:
            raise OSError(
                f"The PCAN-Basic API could not be loaded. ({lib_path})"
            ) from None

    # Initializes a PCAN Channel
    #
    def Initialize(
        self,
        Channel,
        Btr0Btr1,
        HwType=TPCANType(0),
        IOPort=c_uint(0),
        Interrupt=c_ushort(0),
    ):
        """Initializes a PCAN Channel

        Parameters:
          Channel  : A TPCANHandle representing a PCAN Channel
          Btr0Btr1 : The speed for the communication (BTR0BTR1 code)
          HwType   : Non-PnP: The type of hardware and operation mode
          IOPort   : Non-PnP: The I/O address for the parallel port
          Interrupt: Non-PnP: Interrupt number of the parallel port

        Returns:
          A TPCANStatus error code
        """
        try:
            res = self.__m_dllBasic.CAN_Initialize(
                Channel, Btr0Btr1, HwType, IOPort, Interrupt
            )
            return TPCANStatus(res)
        except:
            logger.error("Exception on PCANBasic.Initialize")
            raise

    # Initializes a FD capable PCAN Channel
    #
    def InitializeFD(self, Channel, BitrateFD):
        """Initializes a FD capable PCAN Channel

        Parameters:
          Channel  : The handle of a FD capable PCAN Channel
          BitrateFD : The speed for the communication (FD bit rate string)

        Remarks:
          * See PCAN_BR_* values.
          * parameter and values must be separated by '='
          * Couples of Parameter/value must be separated by ','
          * Following Parameter must be filled out: f_clock, data_brp, data_sjw, data_tseg1, data_tseg2,
            nom_brp, nom_sjw, nom_tseg1, nom_tseg2.
          * Following Parameters are optional (not used yet): data_ssp_offset, nom_sam

        Example:
          f_clock=80000000,nom_brp=10,nom_tseg1=5,nom_tseg2=2,nom_sjw=1,data_brp=4,data_tseg1=7,data_tseg2=2,data_sjw=1

        Returns:
          A TPCANStatus error code
        """
        try:
            res = self.__m_dllBasic.CAN_InitializeFD(Channel, BitrateFD)
            return TPCANStatus(res)
        except:
            logger.error("Exception on PCANBasic.InitializeFD")
            raise

    #  Uninitializes one or all PCAN Channels initialized by CAN_Initialize
    #
    def Uninitialize(self, Channel):
        """Uninitializes one or all PCAN Channels initialized by CAN_Initialize

        Remarks:
          Giving the TPCANHandle value "PCAN_NONEBUS", uninitialize all initialized channels

        Parameters:
          Channel  : A TPCANHandle representing a PCAN Channel

        Returns:
          A TPCANStatus error code
        """
        try:
            res = self.__m_dllBasic.CAN_Uninitialize(Channel)
            return TPCANStatus(res)
        except:
            logger.error("Exception on PCANBasic.Uninitialize")
            raise

    #  Resets the receive and transmit queues of the PCAN Channel
    #
    def Reset(self, Channel):
        """Resets the receive and transmit queues of the PCAN Channel

        Remarks:
          A reset of the CAN controller is not performed

        Parameters:
          Channel  : A TPCANHandle representing a PCAN Channel

        Returns:
          A TPCANStatus error code
        """
        try:
            res = self.__m_dllBasic.CAN_Reset(Channel)
            return TPCANStatus(res)
        except:
            logger.error("Exception on PCANBasic.Reset")
            raise

    #  Gets the current status of a PCAN Channel
    #
    def GetStatus(self, Channel):
        """Gets the current status of a PCAN Channel

        Parameters:
          Channel  : A TPCANHandle representing a PCAN Channel

        Returns:
          A TPCANStatus error code
        """
        try:
            res = self.__m_dllBasic.CAN_GetStatus(Channel)
            return TPCANStatus(res)
        except:
            logger.error("Exception on PCANBasic.GetStatus")
            raise

    # Reads a CAN message from the receive queue of a PCAN Channel
    #
    def Read(self, Channel):
        """Reads a CAN message from the receive queue of a PCAN Channel

        Remarks:
          The return value of this method is a 3-tuple, where
          the first value is the result (TPCANStatus) of the method.
          The order of the values are:
          [0]: A TPCANStatus error code
          [1]: A TPCANMsg structure with the CAN message read
          [2]: A TPCANTimestamp structure with the time when a message was read

        Parameters:
          Channel  : A TPCANHandle representing a PCAN Channel

        Returns:
          A tuple with three values
        """
        try:
            msg = TPCANMsg()
            timestamp = TPCANTimestamp()
            res = self.__m_dllBasic.CAN_Read(Channel, byref(msg), byref(timestamp))
            return TPCANStatus(res), msg, timestamp
        except:
            logger.error("Exception on PCANBasic.Read")
            raise

    # Reads a CAN message from the receive queue of a FD capable PCAN Channel
    #
    def ReadFD(self, Channel):
        """Reads a CAN message from the receive queue of a FD capable PCAN Channel

        Remarks:
          The return value of this method is a 3-tuple, where
          the first value is the result (TPCANStatus) of the method.
          The order of the values are:
          [0]: A TPCANStatus error code
          [1]: A TPCANMsgFD structure with the CAN message read
          [2]: A TPCANTimestampFD that is the time when a message was read

        Parameters:
          Channel  : The handle of a FD capable PCAN Channel

        Returns:
          A tuple with three values
        """
        try:
            msg = TPCANMsgFD()
            timestamp = TPCANTimestampFD()
            res = self.__m_dllBasic.CAN_ReadFD(Channel, byref(msg), byref(timestamp))
            return TPCANStatus(res), msg, timestamp
        except:
            logger.error("Exception on PCANBasic.ReadFD")
            raise

    # Transmits a CAN message
    #
    def Write(self, Channel, MessageBuffer):
        """Transmits a CAN message

        Parameters:
          Channel      : A TPCANHandle representing a PCAN Channel
          MessageBuffer: A TPCANMsg representing the CAN message to be sent

        Returns:
          A TPCANStatus error code
        """
        try:
            res = self.__m_dllBasic.CAN_Write(Channel, byref(MessageBuffer))
            return TPCANStatus(res)
        except:
            logger.error("Exception on PCANBasic.Write")
            raise

    # Transmits a CAN message over a FD capable PCAN Channel
    #
    def WriteFD(self, Channel, MessageBuffer):
        """Transmits a CAN message over a FD capable PCAN Channel

        Parameters:
          Channel      : The handle of a FD capable PCAN Channel
          MessageBuffer: A TPCANMsgFD buffer with the message to be sent

        Returns:
          A TPCANStatus error code
        """
        try:
            res = self.__m_dllBasic.CAN_WriteFD(Channel, byref(MessageBuffer))
            return TPCANStatus(res)
        except:
            logger.error("Exception on PCANBasic.WriteFD")
            raise

    # Configures the reception filter
    #
    def FilterMessages(self, Channel, FromID, ToID, Mode):
        """Configures the reception filter

        Remarks:
          The message filter will be expanded with every call to this function.
          If it is desired to reset the filter, please use the 'SetValue' function.

        Parameters:
          Channel : A TPCANHandle representing a PCAN Channel
          FromID  : A c_uint value with the lowest CAN ID to be received
          ToID    : A c_uint value with the highest CAN ID to be received
          Mode    : A TPCANMode representing the message type (Standard, 11-bit
                    identifier, or Extended, 29-bit identifier)

        Returns:
          A TPCANStatus error code
        """
        try:
            res = self.__m_dllBasic.CAN_FilterMessages(Channel, FromID, ToID, Mode)
            return TPCANStatus(res)
        except:
            logger.error("Exception on PCANBasic.FilterMessages")
            raise

    # Retrieves a PCAN Channel value
    #
    def GetValue(self, Channel, Parameter):
        """Retrieves a PCAN Channel value

        Remarks:
          Parameters can be present or not according with the kind
          of Hardware (PCAN Channel) being used. If a parameter is not available,
          a PCAN_ERROR_ILLPARAMTYPE error will be returned.

          The return value of this method is a 2-tuple, where
          the first value is the result (TPCANStatus) of the method and
          the second one, the asked value

        Parameters:
          Channel   : A TPCANHandle representing a PCAN Channel
          Parameter : The TPCANParameter parameter to get

        Returns:
          A tuple with 2 values
        """
        try:
            if (
                Parameter == PCAN_API_VERSION
                or Parameter == PCAN_HARDWARE_NAME
                or Parameter == PCAN_CHANNEL_VERSION
                or Parameter == PCAN_LOG_LOCATION
                or Parameter == PCAN_TRACE_LOCATION
                or Parameter == PCAN_BITRATE_INFO_FD
                or Parameter == PCAN_IP_ADDRESS
                or Parameter == PCAN_FIRMWARE_VERSION
                or Parameter == PCAN_DEVICE_PART_NUMBER
            ):
                mybuffer = create_string_buffer(256)

            elif Parameter == PCAN_ATTACHED_CHANNELS:
                res = self.GetValue(Channel, PCAN_ATTACHED_CHANNELS_COUNT)
                if TPCANStatus(res[0]) != PCAN_ERROR_OK:
                    return (TPCANStatus(res[0]),)
                mybuffer = (TPCANChannelInformation * res[1])()

            elif (
                Parameter == PCAN_ACCEPTANCE_FILTER_11BIT
                or PCAN_ACCEPTANCE_FILTER_29BIT
            ):
                mybuffer = c_int64(0)

            else:
                mybuffer = c_int(0)

            res = self.__m_dllBasic.CAN_GetValue(
                Channel, Parameter, byref(mybuffer), sizeof(mybuffer)
            )
            if Parameter == PCAN_ATTACHED_CHANNELS:
                return TPCANStatus(res), mybuffer
            else:
                return TPCANStatus(res), mybuffer.value
        except:
            logger.error("Exception on PCANBasic.GetValue")
            raise

    # Returns a descriptive text of a given TPCANStatus
    # error code, in any desired language
    #
    def SetValue(self, Channel, Parameter, Buffer):
        """Returns a descriptive text of a given TPCANStatus error
        code, in any desired language

        Remarks:
          Parameters can be present or not according with the kind
          of Hardware (PCAN Channel) being used. If a parameter is not available,
          a PCAN_ERROR_ILLPARAMTYPE error will be returned.

        Parameters:
          Channel      : A TPCANHandle representing a PCAN Channel
          Parameter    : The TPCANParameter parameter to set
          Buffer       : Buffer with the value to be set
          BufferLength : Size in bytes of the buffer

        Returns:
          A TPCANStatus error code
        """
        try:
            if (
                Parameter == PCAN_LOG_LOCATION
                or Parameter == PCAN_LOG_TEXT
                or Parameter == PCAN_TRACE_LOCATION
            ):
                mybuffer = create_string_buffer(256)
            elif (
                Parameter == PCAN_ACCEPTANCE_FILTER_11BIT
                or PCAN_ACCEPTANCE_FILTER_29BIT
            ):
                mybuffer = c_int64(0)
            else:
                mybuffer = c_int(0)

            mybuffer.value = Buffer
            res = self.__m_dllBasic.CAN_SetValue(
                Channel, Parameter, byref(mybuffer), sizeof(mybuffer)
            )
            return TPCANStatus(res)
        except:
            logger.error("Exception on PCANBasic.SetValue")
            raise

    def GetErrorText(self, Error, Language=0):
        """Configures or sets a PCAN Channel value

        Remarks:

          The current languages available for translation are:
          Neutral (0x00), German (0x07), English (0x09), Spanish (0x0A),
          Italian (0x10) and French (0x0C)

          The return value of this method is a 2-tuple, where
          the first value is the result (TPCANStatus) of the method and
          the second one, the error text

        Parameters:
          Error    : A TPCANStatus error code
          Language : Indicates a 'Primary language ID' (Default is Neutral(0))

        Returns:
          A tuple with 2 values
        """
        try:
            mybuffer = create_string_buffer(256)
            res = self.__m_dllBasic.CAN_GetErrorText(Error, Language, byref(mybuffer))
            return TPCANStatus(res), mybuffer.value
        except:
            logger.error("Exception on PCANBasic.GetErrorText")
            raise

    def LookUpChannel(self, Parameters):
        """Finds a PCAN-Basic channel that matches with the given parameters

        Remarks:

          The return value of this method is a 2-tuple, where
          the first value is the result (TPCANStatus) of the method and
          the second one a TPCANHandle value

        Parameters:
            Parameters   : A comma separated string contained pairs of parameter-name/value
                           to be matched within a PCAN-Basic channel

        Returns:
          A tuple with 2 values
        """
        try:
            mybuffer = TPCANHandle(0)
            res = self.__m_dllBasic.CAN_LookUpChannel(Parameters, byref(mybuffer))
            return TPCANStatus(res), mybuffer
        except:
            logger.error("Exception on PCANBasic.LookUpChannel")
            raise



#!/usr/bin/env python
#
# Copyright (C) 2018 Kristian Sloth Lauszus.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information
# -------------------
# Kristian Sloth Lauszus
# Web      :  http://www.lauszus.com
# e-mail   :  lauszus@gmail.com

import argparse
import math
import os
import random
import struct
import time
import unittest
from collections import defaultdict
from test.config import IS_CI
from unittest.mock import patch

import pytest

import can
from can.viewer import CanViewer, parse_args

# Allow the curses module to be missing (e.g. on PyPy on Windows)
try:
    import curses

    CURSES_AVAILABLE = True
except ImportError:
    curses = None  # type: ignore
    CURSES_AVAILABLE = False


# noinspection SpellCheckingInspection,PyUnusedLocal
class StdscrDummy:
    def __init__(self):
        self.key_counter = 0
        self.draw_buffer = defaultdict(dict)

    @staticmethod
    def clear():
        pass

    @staticmethod
    def erase():
        pass

    @staticmethod
    def getmaxyx():
        # Set y-value, so scrolling gets tested
        # Set x-value, so the text will fit in the window
        return 1, 100

    def addstr(self, row, col, txt, *args):
        assert row >= 0
        assert col >= 0
        assert txt is not None

        # Save the text written into the buffer
        for i, t in enumerate(txt):
            self.draw_buffer[row][col + i] = t

        # Raise an exception 50 % of the time, so we can make sure the code handles it
        if random.random() < 0.5:
            raise curses.error

    @staticmethod
    def nodelay(_bool):
        pass

    def getch(self):
        assert curses is not None

        self.key_counter += 1
        if self.key_counter == 1:
            # Send invalid key
            return -1
        elif self.key_counter == 2:
            return ord("c")  # Clear
        elif self.key_counter == 3:
            return curses.ascii.SP  # Pause
        elif self.key_counter == 4:
            return curses.ascii.SP  # Unpause
        elif self.key_counter == 5:
            return ord("s")  # Sort
        # Turn on byte highlighting (toggle)
        elif self.key_counter == 6:
            return ord("h")
        # Turn off byte highlighting (toggle)
        elif self.key_counter == 7:
            return ord("h")
        # Keep scrolling until it exceeds the number of messages
        elif self.key_counter <= 100:
            return curses.KEY_DOWN
        # Scroll until the header is back as the first line and then scroll over the limit
        elif self.key_counter <= 200:
            return curses.KEY_UP

        return curses.ascii.ESC


@unittest.skipUnless(CURSES_AVAILABLE, "curses might be missing on some platforms")
class CanViewerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Set seed, so the tests are not affected
        random.seed(0)

    def setUp(self):
        self.stdscr_dummy = StdscrDummy()
        config = {"interface": "virtual", "receive_own_messages": True}
        bus = can.Bus(**config)
        data_structs = None

        patch_curs_set = patch("curses.curs_set")
        patch_curs_set.start()
        self.addCleanup(patch_curs_set.stop)

        patch_use_default_colors = patch("curses.use_default_colors")
        patch_use_default_colors.start()
        self.addCleanup(patch_use_default_colors.stop)

        patch_init_pair = patch("curses.init_pair")
        patch_init_pair.start()
        self.addCleanup(patch_init_pair.stop)

        patch_color_pair = patch("curses.color_pair")
        patch_color_pair.start()
        self.addCleanup(patch_color_pair.stop)

        patch_is_term_resized = patch("curses.is_term_resized")
        mock_is_term_resized = patch_is_term_resized.start()
        mock_is_term_resized.return_value = True if random.random() < 0.5 else False
        self.addCleanup(patch_is_term_resized.stop)

        if hasattr(curses, "resizeterm"):
            patch_resizeterm = patch("curses.resizeterm")
            patch_resizeterm.start()
            self.addCleanup(patch_resizeterm.stop)

        self.can_viewer = CanViewer(self.stdscr_dummy, bus, data_structs, testing=True)

    def tearDown(self):
        # Run the viewer after the test, this is done, so we can receive the CAN-Bus messages and make sure that they
        # are parsed correctly
        self.can_viewer.run()

    def test_send(self):
        # CANopen EMCY
        data = [1, 2, 3, 4, 5, 6, 7]  # Wrong length
        msg = can.Message(arbitration_id=0x080 + 1, data=data, is_extended_id=False)
        self.can_viewer.bus.send(msg)

        data = [1, 2, 3, 4, 5, 6, 7, 8]
        msg = can.Message(arbitration_id=0x080 + 1, data=data, is_extended_id=False)
        self.can_viewer.bus.send(msg)

        # CANopen HEARTBEAT
        data = [0x05]  # Operational
        msg = can.Message(arbitration_id=0x700 + 0x7F, data=data, is_extended_id=False)
        self.can_viewer.bus.send(msg)

        # Send non-CANopen message
        data = [1, 2, 3, 4, 5, 6, 7, 8]
        msg = can.Message(arbitration_id=0x101, data=data, is_extended_id=False)
        self.can_viewer.bus.send(msg)

        # Send the same command, but with another data length
        data = [1, 2, 3, 4, 5, 6]
        msg = can.Message(arbitration_id=0x101, data=data, is_extended_id=False)
        self.can_viewer.bus.send(msg)

        # Send non-CANopen message with long parsed data length
        data = [255, 255]
        msg = can.Message(arbitration_id=0x102, data=data, is_extended_id=False)
        self.can_viewer.bus.send(msg)

        # Send the same command, but with shorter parsed data length
        data = [0, 0]
        msg = can.Message(arbitration_id=0x102, data=data, is_extended_id=False)
        self.can_viewer.bus.send(msg)

        # Message with extended id
        data = [1, 2, 3, 4, 5, 6, 7, 8]
        msg = can.Message(arbitration_id=0x123456, data=data, is_extended_id=True)
        self.can_viewer.bus.send(msg)
        # self.assertTupleEqual(self.can_viewer.parse_canopen_message(msg), (None, None))

        # Send the same message again to make sure that resending works and dt is correct
        time.sleep(0.1)
        self.can_viewer.bus.send(msg)

        # Send error message
        msg = can.Message(is_error_frame=True)
        self.can_viewer.bus.send(msg)

    def test_receive(self):
        # Send the messages again, but this time the test code will receive it
        self.test_send()

        data_structs = {
            # For converting the EMCY and HEARTBEAT messages
            0x080 + 0x01: struct.Struct("<HBLB"),
            0x700 + 0x7F: struct.Struct("<B"),
            # Shorter parsed data length
            0x102: struct.Struct("<BB"),
            # Big-endian and float test
            0x123456: struct.Struct(">ff"),
        }
        # Receive the messages we just sent in 'test_canopen'
        while 1:
            msg = self.can_viewer.bus.recv(timeout=0)
            if msg is not None:
                self.can_viewer.data_structs = (
                    data_structs if msg.arbitration_id != 0x101 else None
                )
                _id = self.can_viewer.draw_can_bus_message(msg)
                if _id["msg"].arbitration_id == 0x101:
                    # Check if the counter is reset when the length has changed
                    self.assertEqual(_id["count"], 1)

                    # Make sure the line has been cleared after the shorted message was send
                    for col, v in self.stdscr_dummy.draw_buffer[_id["row"]].items():
                        if col >= 52 + _id["msg"].dlc * 3:
                            self.assertEqual(v, " ")
                elif _id["msg"].arbitration_id == 0x102:
                    # Make sure the parsed values have been cleared after the shorted message was send
                    for col, v in self.stdscr_dummy.draw_buffer[_id["row"]].items():
                        if col >= 77 + _id["values_string_length"]:
                            self.assertEqual(v, " ")
                elif _id["msg"].arbitration_id == 0x123456:
                    # Check if the counter is incremented
                    if _id["dt"] == 0:
                        self.assertEqual(_id["count"], 1)
                    else:
                        if not IS_CI:  # do not test timing in CI
                            assert _id["dt"] == pytest.approx(
                                0.1, abs=5e-2
                            )  # dt should be ~0.1 s
                        self.assertEqual(_id["count"], 2)
                else:
                    # Make sure dt is 0
                    if _id["count"] == 1:
                        self.assertEqual(_id["dt"], 0)
            else:
                break

    # Convert it into raw integer values and then pack the data
    @staticmethod
    def pack_data(
        cmd, cmd_to_struct, *args
    ):  # type: (int, Dict, Union[*float, *int]) -> bytes
        if not cmd_to_struct or len(args) == 0:
            # If no arguments are given, then the message does not contain a data package
            return b""

        for key in cmd_to_struct.keys():
            if cmd == key if isinstance(key, int) else cmd in key:
                value = cmd_to_struct[key]
                if isinstance(value, tuple):
                    # The struct is given as the fist argument
                    struct_t = value[0]  # type: struct.Struct

                    # The conversion from SI-units to raw values are given in the rest of the tuple
                    fmt = struct_t.format
                    if isinstance(fmt, str):  # pragma: no cover
                        # Needed for Python 3.7
                        fmt = fmt.encode()

                    # Make sure the endian is given as the first argument
                    assert fmt[0] == ord("<") or fmt[0] == ord(">")

                    # Disable rounding if the format is a float
                    data = []
                    for c, arg, val in zip(fmt[1:], args, value[1:]):
                        if c == ord("f"):
                            data.append(arg * val)
                        else:
                            data.append(round(arg * val))
                else:
                    # No conversion from SI-units is needed
                    struct_t = value  # type: struct.Struct
                    data = args

                return struct_t.pack(*data)
        else:
            raise ValueError(f"Unknown command: 0x{cmd:02X}")

    def test_pack_unpack(self):
        CANOPEN_TPDO1 = 0x180
        CANOPEN_TPDO2 = 0x280
        CANOPEN_TPDO3 = 0x380
        CANOPEN_TPDO4 = 0x480

        # Dictionary used to convert between Python values and C structs represented as Python strings.
        # If the value is 'None' then the message does not contain any data package.
        #
        # The struct package is used to unpack the received data.
        # Note the data is assumed to be in little-endian byte order.
        # < = little-endian, > = big-endian
        # x = pad byte
        # c = char
        # ? = bool
        # b = int8_t, B = uint8_t
        # h = int16, H = uint16
        # l = int32_t, L = uint32_t
        # q = int64_t, Q = uint64_t
        # f = float (32-bits), d = double (64-bits)
        #
        # An optional conversion from real units to integers can be given as additional arguments.
        # In order to convert from raw integer value the SI-units are multiplied with the values and similarly the values
        # are divided by the value in order to convert from real units to raw integer values.
        data_structs = {
            # CANopen node 1
            CANOPEN_TPDO1 + 1: struct.Struct("<bBh2H"),
            CANOPEN_TPDO2 + 1: (struct.Struct("<HHB"), 100.0, 10.0, 1),
            CANOPEN_TPDO3 + 1: struct.Struct("<ff"),
            CANOPEN_TPDO4 + 1: (struct.Struct("<ff"), math.pi / 180.0, math.pi / 180.0),
            CANOPEN_TPDO1 + 2: None,
            CANOPEN_TPDO2 + 2: struct.Struct(">lL"),
            (CANOPEN_TPDO3 + 2, CANOPEN_TPDO4 + 2): struct.Struct(">LL"),
        }  # type: Dict[Union[int, Tuple[int, ...]], Union[struct.Struct, Tuple, None]]

        raw_data = self.pack_data(
            CANOPEN_TPDO1 + 1, data_structs, -7, 13, -1024, 2048, 0xFFFF
        )
        parsed_data = CanViewer.unpack_data(CANOPEN_TPDO1 + 1, data_structs, raw_data)
        self.assertListEqual(parsed_data, [-7, 13, -1024, 2048, 0xFFFF])
        self.assertTrue(all(isinstance(d, int) for d in parsed_data))

        raw_data = self.pack_data(CANOPEN_TPDO2 + 1, data_structs, 12.34, 4.5, 6)
        parsed_data = CanViewer.unpack_data(CANOPEN_TPDO2 + 1, data_structs, raw_data)
        assert parsed_data == pytest.approx([12.34, 4.5, 6])
        self.assertTrue(
            isinstance(parsed_data[0], float)
            and isinstance(parsed_data[1], float)
            and isinstance(parsed_data[2], int)
        )

        raw_data = self.pack_data(CANOPEN_TPDO3 + 1, data_structs, 123.45, 67.89)
        parsed_data = CanViewer.unpack_data(CANOPEN_TPDO3 + 1, data_structs, raw_data)
        assert parsed_data == pytest.approx([123.45, 67.89])
        self.assertTrue(all(isinstance(d, float) for d in parsed_data))

        raw_data = self.pack_data(
            CANOPEN_TPDO4 + 1, data_structs, math.pi / 2.0, math.pi
        )
        parsed_data = CanViewer.unpack_data(CANOPEN_TPDO4 + 1, data_structs, raw_data)
        assert parsed_data == pytest.approx([math.pi / 2.0, math.pi])
        self.assertTrue(all(isinstance(d, float) for d in parsed_data))

        raw_data = self.pack_data(CANOPEN_TPDO1 + 2, data_structs)
        parsed_data = CanViewer.unpack_data(CANOPEN_TPDO1 + 2, data_structs, raw_data)
        self.assertListEqual(parsed_data, [])
        self.assertIsInstance(parsed_data, list)

        raw_data = self.pack_data(
            CANOPEN_TPDO2 + 2, data_structs, -2147483648, 0xFFFFFFFF
        )
        parsed_data = CanViewer.unpack_data(CANOPEN_TPDO2 + 2, data_structs, raw_data)
        self.assertListEqual(parsed_data, [-2147483648, 0xFFFFFFFF])

        raw_data = self.pack_data(CANOPEN_TPDO3 + 2, data_structs, 0xFF, 0xFFFF)
        parsed_data = CanViewer.unpack_data(CANOPEN_TPDO3 + 2, data_structs, raw_data)
        self.assertListEqual(parsed_data, [0xFF, 0xFFFF])

        raw_data = self.pack_data(CANOPEN_TPDO4 + 2, data_structs, 0xFFFFFF, 0xFFFFFFFF)
        parsed_data = CanViewer.unpack_data(CANOPEN_TPDO4 + 2, data_structs, raw_data)
        self.assertListEqual(parsed_data, [0xFFFFFF, 0xFFFFFFFF])

        self.assertTrue(all(isinstance(d, int) for d in parsed_data))

        # Make sure that the ValueError exception is raised
        with self.assertRaises(ValueError):
            self.pack_data(0x101, data_structs, 1, 2, 3, 4)

        with self.assertRaises(ValueError):
            CanViewer.unpack_data(
                0x102, data_structs, b"\x01\x02\x03\x04\x05\x06\x07\x08"
            )

    def test_parse_args(self):
        parsed_args, _, _, _ = parse_args(["-b", "250000"])
        self.assertEqual(parsed_args.bitrate, 250000)

        parsed_args, _, _, _ = parse_args(["--bitrate", "500000"])
        self.assertEqual(parsed_args.bitrate, 500000)

        parsed_args, _, _, _ = parse_args(["-c", "can0"])
        self.assertEqual(parsed_args.channel, "can0")

        parsed_args, _, _, _ = parse_args(["--channel", "PCAN_USBBUS1"])
        self.assertEqual(parsed_args.channel, "PCAN_USBBUS1")

        parsed_args, _, data_structs, _ = parse_args(["-d", "100:<L"])
        self.assertEqual(parsed_args.decode, ["100:<L"])

        self.assertIsInstance(data_structs, dict)
        self.assertEqual(len(data_structs), 1)

        self.assertIsInstance(data_structs[0x100], struct.Struct)
        self.assertIn(data_structs[0x100].format, ["<L", b"<L"])
        self.assertEqual(data_structs[0x100].size, 4)

        f = open("test.txt", "w")
        f.write("100:<BB\n101:<HH\n")
        f.close()
        parsed_args, _, data_structs, _ = parse_args(["-d", "test.txt"])

        self.assertIsInstance(data_structs, dict)
        self.assertEqual(len(data_structs), 2)

        self.assertIsInstance(data_structs[0x100], struct.Struct)
        self.assertIn(data_structs[0x100].format, ["<BB", b"<BB"])
        self.assertEqual(data_structs[0x100].size, 2)

        self.assertIsInstance(data_structs[0x101], struct.Struct)
        self.assertIn(data_structs[0x101].format, ["<HH", b"<HH"])
        self.assertEqual(data_structs[0x101].size, 4)
        os.remove("test.txt")

        parsed_args, _, data_structs, _ = parse_args(
            ["--decode", "100:<LH:10.:100.", "101:<ff", "102:<Bf:1:57.3"]
        )
        self.assertEqual(
            parsed_args.decode, ["100:<LH:10.:100.", "101:<ff", "102:<Bf:1:57.3"]
        )

        self.assertIsInstance(data_structs, dict)
        self.assertEqual(len(data_structs), 3)

        self.assertIsInstance(data_structs[0x100], tuple)
        self.assertEqual(len(data_structs[0x100]), 3)

        self.assertIsInstance(data_structs[0x100][0], struct.Struct)
        self.assertIsInstance(data_structs[0x100][1], float)
        self.assertIsInstance(data_structs[0x100][2], float)
        self.assertIn(data_structs[0x100][0].format, ["<LH", b"<LH"])
        self.assertEqual(data_structs[0x100][0].size, 6)
        self.assertEqual(data_structs[0x100][1], 10.0)
        self.assertEqual(data_structs[0x100][2], 100.0)

        self.assertIsInstance(data_structs[0x101], struct.Struct)
        self.assertIn(data_structs[0x101].format, ["<ff", b"<ff"])
        self.assertEqual(data_structs[0x101].size, 8)

        self.assertIsInstance(data_structs[0x102][0], struct.Struct)
        self.assertIsInstance(data_structs[0x102][1], int)
        self.assertIsInstance(data_structs[0x102][2], float)
        self.assertIn(data_structs[0x102][0].format, ["<Bf", b"<Bf"])
        self.assertEqual(data_structs[0x102][0].size, 5)
        self.assertEqual(data_structs[0x102][1], 1)
        self.assertAlmostEqual(data_structs[0x102][2], 57.3)

        parsed_args, can_filters, _, _ = parse_args(["-f", "100:7FF"])
        self.assertEqual(parsed_args.filter, ["100:7FF"])
        self.assertIsInstance(can_filters, list)
        self.assertIsInstance(can_filters[0], dict)
        self.assertEqual(can_filters[0]["can_id"], 0x100)
        self.assertEqual(can_filters[0]["can_mask"], 0x7FF)

        parsed_args, can_filters, _, _ = parse_args(["-f", "101:7FF", "102:7FC"])
        self.assertEqual(parsed_args.filter, ["101:7FF", "102:7FC"])
        self.assertIsInstance(can_filters, list)
        self.assertIsInstance(can_filters[0], dict)
        self.assertIsInstance(can_filters[1], dict)
        self.assertEqual(can_filters[0]["can_id"], 0x101)
        self.assertEqual(can_filters[0]["can_mask"], 0x7FF)
        self.assertEqual(can_filters[1]["can_id"], 0x102)
        self.assertEqual(can_filters[1]["can_mask"], 0x7FC)

        with self.assertRaises(argparse.ArgumentError):
            parse_args(["-f", "101,7FF"])

        parsed_args, can_filters, _, _ = parse_args(["--filter", "100~7FF"])
        self.assertEqual(parsed_args.filter, ["100~7FF"])
        self.assertIsInstance(can_filters, list)
        self.assertIsInstance(can_filters[0], dict)
        self.assertEqual(can_filters[0]["can_id"], 0x100 | 0x20000000)
        self.assertEqual(can_filters[0]["can_mask"], 0x7FF & 0x20000000)

        parsed_args, _, _, _ = parse_args(["-i", "socketcan"])
        self.assertEqual(parsed_args.interface, "socketcan")

        parsed_args, _, _, _ = parse_args(["--interface", "pcan"])
        self.assertEqual(parsed_args.interface, "pcan")

        # Make sure it exits with the correct error code when displaying the help page
        # See: https://github.com/hardbyte/python-can/issues/427
        with self.assertRaises(SystemExit) as cm:
            parse_args(["-h"])
        self.assertEqual(cm.exception.code, 0)

        with self.assertRaises(SystemExit) as cm:
            parse_args([])
        import errno

        self.assertEqual(cm.exception.code, errno.EINVAL)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

"""
This test module test the separate reader/writer combinations of the can.io.*
modules by writing some messages to a temporary file and reading it again.
Then it checks if the messages that were read are same ones as the
ones that were written. It also checks that the order of the messages
is correct. The types of messages that are tested differs between the
different writer/reader pairs - e.g., some don't handle error frames and
comments.

TODO: correctly set preserves_channel and adds_default_channel
"""
import logging
import os
import tempfile
import unittest
from abc import ABCMeta, abstractmethod
from datetime import datetime
from itertools import zip_longest

from parameterized import parameterized

import can
from can.io import blf

from .data.example_data import (
    TEST_COMMENTS,
    TEST_MESSAGES_BASE,
    TEST_MESSAGES_CAN_FD,
    TEST_MESSAGES_ERROR_FRAMES,
    TEST_MESSAGES_REMOTE_FRAMES,
    sort_messages,
)
from .message_helper import ComparingMessagesTestCase

logging.basicConfig(level=logging.DEBUG)

try:
    import asammdf
except ModuleNotFoundError:
    asammdf = None


class ReaderWriterExtensionTest(unittest.TestCase):
    def _get_suffix_case_variants(self, suffix):
        return [
            suffix.upper(),
            suffix.lower(),
            f"can.msg.ext{suffix}",
            "".join([c.upper() if i % 2 else c for i, c in enumerate(suffix)]),
        ]

    def _test_extension(self, suffix):
        WriterType = can.Logger.message_writers.get(suffix)
        ReaderType = can.LogReader.message_readers.get(suffix)
        for suffix_variant in self._get_suffix_case_variants(suffix):
            tmp_file = tempfile.NamedTemporaryFile(suffix=suffix_variant, delete=False)
            tmp_file.close()
            try:
                if WriterType:
                    with can.Logger(tmp_file.name) as logger:
                        assert type(logger) == WriterType
                if ReaderType:
                    with can.LogReader(tmp_file.name) as player:
                        assert type(player) == ReaderType
            finally:
                os.remove(tmp_file.name)

    def test_extension_matching_asc(self):
        self._test_extension(".asc")

    def test_extension_matching_blf(self):
        self._test_extension(".blf")

    def test_extension_matching_csv(self):
        self._test_extension(".csv")

    def test_extension_matching_db(self):
        self._test_extension(".db")

    def test_extension_matching_log(self):
        self._test_extension(".log")

    def test_extension_matching_txt(self):
        self._test_extension(".txt")

    def test_extension_matching_mf4(self):
        try:
            self._test_extension(".mf4")
        except NotImplementedError:
            if asammdf is not None:
                raise


class ReaderWriterTest(unittest.TestCase, ComparingMessagesTestCase, metaclass=ABCMeta):
    """Tests a pair of writer and reader by writing all data first and
    then reading all data and checking if they could be reconstructed
    correctly. Optionally writes some comments as well.

    .. note::
        This class is prevented from being executed as a test
        case itself by a *del* statement in at the end of the file.
        (Source: `*Wojciech B.* on StackOverlfow <https://stackoverflow.com/a/22836015/3753684>`_)
    """

    def __init__(self, *args, **kwargs):
        unittest.TestCase.__init__(self, *args, **kwargs)
        self._setup_instance()

    @abstractmethod
    def _setup_instance(self):
        """Hook for subclasses."""
        raise NotImplementedError()

    def _setup_instance_helper(
        self,
        writer_constructor,
        reader_constructor,
        binary_file=False,
        check_remote_frames=True,
        check_error_frames=True,
        check_fd=True,
        check_comments=False,
        test_append=False,
        allowed_timestamp_delta=0.0,
        preserves_channel=True,
        adds_default_channel=None,
    ):
        """
        :param Callable writer_constructor: the constructor of the writer class
        :param Callable reader_constructor: the constructor of the reader class
        :param bool binary_file: if True, opens files in binary and not in text mode

        :param bool check_remote_frames: if True, also tests remote frames
        :param bool check_error_frames: if True, also tests error frames
        :param bool check_fd: if True, also tests CAN FD frames
        :param bool check_comments: if True, also inserts comments at some
                                    locations and checks if they are contained anywhere literally
                                    in the resulting file. The locations as selected randomly
                                    but deterministically, which makes the test reproducible.
        :param bool test_append: tests the writer in append mode as well

        :param float or int or None allowed_timestamp_delta: directly passed to :meth:`can.Message.equals`
        :param bool preserves_channel: if True, checks that the channel attribute is preserved
        :param any adds_default_channel: sets this as the channel when not other channel was given
                                         ignored, if *preserves_channel* is True
        """
        # get all test messages
        self.original_messages = list(TEST_MESSAGES_BASE)
        if check_remote_frames:
            self.original_messages += TEST_MESSAGES_REMOTE_FRAMES
        if check_error_frames:
            self.original_messages += TEST_MESSAGES_ERROR_FRAMES
        if check_fd:
            self.original_messages += TEST_MESSAGES_CAN_FD

        # sort them so that for example ASCWriter does not "fix" any messages with timestamp 0.0
        self.original_messages = sort_messages(self.original_messages)

        if check_comments:
            # we check this because of the lack of a common base class
            # we filter for not starts with '__' so we do not get all the builtin
            # methods when logging to the console
            attrs = [
                attr for attr in dir(writer_constructor) if not attr.startswith("__")
            ]
            assert (
                "log_event" in attrs
            ), f"cannot check comments with this writer: {writer_constructor}"

        # get all test comments
        self.original_comments = TEST_COMMENTS if check_comments else ()

        self.writer_constructor = writer_constructor
        self.reader_constructor = reader_constructor
        self.binary_file = binary_file
        self.test_append_enabled = test_append

        ComparingMessagesTestCase.__init__(
            self,
            allowed_timestamp_delta=allowed_timestamp_delta,
            preserves_channel=preserves_channel,
        )
        # adds_default_channel=adds_default_channel # TODO inlcude in tests

    def setUp(self):
        with tempfile.NamedTemporaryFile("w+", delete=False) as test_file:
            self.test_file_name = test_file.name

    def tearDown(self):
        os.remove(self.test_file_name)
        del self.test_file_name

    def test_path_like_explicit_stop(self):
        """testing with path-like and explicit stop() call"""

        # create writer
        print("writing all messages/comments")
        writer = self.writer_constructor(self.test_file_name)
        self._write_all(writer)
        self._ensure_fsync(writer)
        writer.stop()
        if hasattr(writer.file, "closed"):
            self.assertTrue(writer.file.closed)

        print("reading all messages")
        reader = self.reader_constructor(self.test_file_name)
        read_messages = list(reader)
        # redundant, but this checks if stop() can be called multiple times
        reader.stop()
        if hasattr(writer.file, "closed"):
            self.assertTrue(writer.file.closed)

        # check if at least the number of messages matches
        # could use assertCountEqual in later versions of Python and in the other methods
        self.assertEqual(
            len(read_messages),
            len(self.original_messages),
            "the number of written messages does not match the number of read messages",
        )

        self.assertMessagesEqual(self.original_messages, read_messages)
        self.assertIncludesComments(self.test_file_name)

    def test_path_like_context_manager(self):
        """testing with path-like object and context manager"""

        # create writer
        print("writing all messages/comments")
        with self.writer_constructor(self.test_file_name) as writer:
            self._write_all(writer)
            self._ensure_fsync(writer)
            w = writer
        if hasattr(w.file, "closed"):
            self.assertTrue(w.file.closed)

        # read all written messages
        print("reading all messages")
        with self.reader_constructor(self.test_file_name) as reader:
            read_messages = list(reader)
            r = reader
        if hasattr(r.file, "closed"):
            self.assertTrue(r.file.closed)

        # check if at least the number of messages matches;
        self.assertEqual(
            len(read_messages),
            len(self.original_messages),
            "the number of written messages does not match the number of read messages",
        )

        self.assertMessagesEqual(self.original_messages, read_messages)
        self.assertIncludesComments(self.test_file_name)

    def test_file_like_explicit_stop(self):
        """testing with file-like object and explicit stop() call"""

        # create writer
        print("writing all messages/comments")
        my_file = open(self.test_file_name, "wb" if self.binary_file else "w")
        writer = self.writer_constructor(my_file)
        self._write_all(writer)
        self._ensure_fsync(writer)
        writer.stop()
        if hasattr(my_file, "closed"):
            self.assertTrue(my_file.closed)

        print("reading all messages")
        my_file = open(self.test_file_name, "rb" if self.binary_file else "r")
        reader = self.reader_constructor(my_file)
        read_messages = list(reader)
        # redundant, but this checks if stop() can be called multiple times
        reader.stop()
        if hasattr(my_file, "closed"):
            self.assertTrue(my_file.closed)

        # check if at least the number of messages matches
        # could use assertCountEqual in later versions of Python and in the other methods
        self.assertEqual(
            len(read_messages),
            len(self.original_messages),
            "the number of written messages does not match the number of read messages",
        )

        self.assertMessagesEqual(self.original_messages, read_messages)
        self.assertIncludesComments(self.test_file_name)

    def test_file_like_context_manager(self):
        """testing with file-like object and context manager"""

        # create writer
        print("writing all messages/comments")
        my_file = open(self.test_file_name, "wb" if self.binary_file else "w")
        with self.writer_constructor(my_file) as writer:
            self._write_all(writer)
            self._ensure_fsync(writer)
            w = writer
        if hasattr(my_file, "closed"):
            self.assertTrue(my_file.closed)

        # read all written messages
        print("reading all messages")
        my_file = open(self.test_file_name, "rb" if self.binary_file else "r")
        with self.reader_constructor(my_file) as reader:
            read_messages = list(reader)
            r = reader
        if hasattr(my_file, "closed"):
            self.assertTrue(my_file.closed)

        # check if at least the number of messages matches;
        self.assertEqual(
            len(read_messages),
            len(self.original_messages),
            "the number of written messages does not match the number of read messages",
        )

        self.assertMessagesEqual(self.original_messages, read_messages)
        self.assertIncludesComments(self.test_file_name)

    def test_append_mode(self):
        """
        testing append mode with context manager and path-like object
        """
        if not self.test_append_enabled:
            raise unittest.SkipTest("do not test append mode")

        count = len(self.original_messages)
        first_part = self.original_messages[: count // 2]
        second_part = self.original_messages[count // 2 :]

        # write first half
        with self.writer_constructor(self.test_file_name) as writer:
            for message in first_part:
                writer(message)
            self._ensure_fsync(writer)

        # use append mode for second half
        try:
            writer = self.writer_constructor(self.test_file_name, append=True)
        except ValueError as e:
            # maybe "append" is not a formal parameter (this is the case for SqliteWriter)
            try:
                writer = self.writer_constructor(self.test_file_name)
            except TypeError:
                # if it is still a problem, raise the initial error
                raise e
        with writer:
            for message in second_part:
                writer(message)
            self._ensure_fsync(writer)
        with self.reader_constructor(self.test_file_name) as reader:
            read_messages = list(reader)

        self.assertMessagesEqual(self.original_messages, read_messages)

    def _write_all(self, writer):
        """Writes messages and insert comments here and there."""
        # Note: we make no assumptions about the length of original_messages and original_comments
        for msg, comment in zip_longest(
            self.original_messages, self.original_comments, fillvalue=None
        ):
            # msg and comment might be None
            if comment is not None:
                print("writing comment: ", comment)
                writer.log_event(comment)  # we already know that this method exists
            if msg is not None:
                print("writing message: ", msg)
                writer(msg)

    def _ensure_fsync(self, io_handler):
        if hasattr(io_handler.file, "fileno"):
            io_handler.file.flush()
            os.fsync(io_handler.file.fileno())

    def assertIncludesComments(self, filename):
        """
        Ensures that all comments are literally contained in the given file.

        :param filename: the path-like object to use
        """
        if self.original_comments:
            # read the entire outout file
            with open(filename, "rb" if self.binary_file else "r") as file:
                output_contents = file.read()
            # check each, if they can be found in there literally
            for comment in self.original_comments:
                self.assertIn(comment, output_contents)


class TestAscFileFormat(ReaderWriterTest):
    """Tests can.ASCWriter and can.ASCReader"""

    FORMAT_START_OF_FILE_DATE = "%a %b %d %I:%M:%S.%f %p %Y"

    def _setup_instance(self):
        super()._setup_instance_helper(
            can.ASCWriter,
            can.ASCReader,
            check_fd=True,
            check_comments=True,
            preserves_channel=False,
            adds_default_channel=0,
        )

    def _read_log_file(self, filename, **kwargs):
        logfile = os.path.join(os.path.dirname(__file__), "data", filename)
        with can.ASCReader(logfile, **kwargs) as reader:
            return list(reader)

    def test_absolute_time(self):
        time_from_file = "Sat Sep 30 10:06:13.191 PM 2017"
        start_time = datetime.strptime(
            time_from_file, self.FORMAT_START_OF_FILE_DATE
        ).timestamp()

        expected_messages = [
            can.Message(
                timestamp=2.5010 + start_time,
                arbitration_id=0xC8,
                is_extended_id=False,
                is_rx=False,
                channel=1,
                dlc=8,
                data=[9, 8, 7, 6, 5, 4, 3, 2],
            ),
            can.Message(
                timestamp=17.876708 + start_time,
                arbitration_id=0x6F9,
                is_extended_id=False,
                channel=0,
                dlc=0x8,
                data=[5, 0xC, 0, 0, 0, 0, 0, 0],
            ),
        ]
        actual = self._read_log_file("test_CanMessage.asc", relative_timestamp=False)
        self.assertMessagesEqual(actual, expected_messages)

    def test_can_message(self):
        expected_messages = [
            can.Message(
                timestamp=2.5010,
                arbitration_id=0xC8,
                is_extended_id=False,
                is_rx=False,
                channel=1,
                dlc=8,
                data=[9, 8, 7, 6, 5, 4, 3, 2],
            ),
            can.Message(
                timestamp=17.876708,
                arbitration_id=0x6F9,
                is_extended_id=False,
                channel=0,
                dlc=0x8,
                data=[5, 0xC, 0, 0, 0, 0, 0, 0],
            ),
        ]
        actual = self._read_log_file("test_CanMessage.asc")
        self.assertMessagesEqual(actual, expected_messages)

    def test_can_remote_message(self):
        expected_messages = [
            can.Message(
                timestamp=2.510001,
                arbitration_id=0x100,
                is_extended_id=False,
                channel=1,
                is_remote_frame=True,
            ),
            can.Message(
                timestamp=2.520002,
                arbitration_id=0x200,
                is_extended_id=False,
                is_rx=False,
                channel=2,
                is_remote_frame=True,
            ),
            can.Message(
                timestamp=2.584921,
                arbitration_id=0x300,
                is_extended_id=False,
                channel=3,
                dlc=8,
                is_remote_frame=True,
            ),
        ]
        actual = self._read_log_file("test_CanRemoteMessage.asc")
        self.assertMessagesEqual(actual, expected_messages)

    def test_can_fd_remote_message(self):
        expected_messages = [
            can.Message(
                timestamp=30.300981,
                arbitration_id=0x50005,
                channel=2,
                dlc=5,
                is_rx=False,
                is_fd=True,
                is_remote_frame=True,
                error_state_indicator=True,
            )
        ]
        actual = self._read_log_file("test_CanFdRemoteMessage.asc")
        self.assertMessagesEqual(actual, expected_messages)

    def test_can_fd_message(self):
        expected_messages = [
            can.Message(
                timestamp=30.005021,
                arbitration_id=0x300,
                is_extended_id=False,
                channel=0,
                dlc=8,
                data=[0x11, 0xC2, 3, 4, 5, 6, 7, 8],
                is_fd=True,
                bitrate_switch=True,
            ),
            can.Message(
                timestamp=30.005041,
                arbitration_id=0x1C4D80A7,
                channel=1,
                dlc=8,
                is_rx=False,
                data=[0x12, 0xC2, 3, 4, 5, 6, 7, 8],
                is_fd=True,
                error_state_indicator=True,
            ),
            can.Message(
                timestamp=30.005071,
                arbitration_id=0x30A,
                is_extended_id=False,
                channel=2,
                dlc=8,
                data=[1, 2, 3, 4, 5, 6, 7, 8],
                is_fd=True,
                bitrate_switch=True,
                error_state_indicator=True,
            ),
        ]
        actual = self._read_log_file("test_CanFdMessage.asc")
        self.assertMessagesEqual(actual, expected_messages)

    def test_can_fd_message_64(self):
        expected_messages = [
            can.Message(
                timestamp=30.506898,
                arbitration_id=0x4EE,
                is_extended_id=False,
                channel=3,
                dlc=64,
                data=[0xA1, 2, 3, 4] + 59 * [0] + [0x64],
                is_fd=True,
                error_state_indicator=True,
            ),
            can.Message(
                timestamp=31.506898,
                arbitration_id=0x1C4D80A7,
                channel=3,
                dlc=64,
                data=[0xB1, 2, 3, 4] + 59 * [0] + [0x64],
                is_fd=True,
                bitrate_switch=True,
            ),
        ]
        actual = self._read_log_file("test_CanFdMessage64.asc")
        self.assertMessagesEqual(actual, expected_messages)

    def test_can_and_canfd_error_frames(self):
        expected_messages = [
            can.Message(timestamp=2.501000, channel=0, is_error_frame=True),
            can.Message(timestamp=3.501000, channel=0, is_error_frame=True),
            can.Message(timestamp=4.501000, channel=1, is_error_frame=True),
            can.Message(
                timestamp=30.806898,
                channel=4,
                is_rx=False,
                is_error_frame=True,
                is_fd=True,
            ),
        ]
        actual = self._read_log_file("test_CanErrorFrames.asc")
        self.assertMessagesEqual(actual, expected_messages)

    def test_ignore_comments(self):
        _msg_list = self._read_log_file("logfile.asc")

    def test_no_triggerblock(self):
        _msg_list = self._read_log_file("issue_1256.asc")

    def test_can_dlc_greater_than_8(self):
        _msg_list = self._read_log_file("issue_1299.asc")


class TestBlfFileFormat(ReaderWriterTest):
    """Tests can.BLFWriter and can.BLFReader.

    Uses log files created by Toby Lorenz:
    https://bitbucket.org/tobylorenz/vector_blf/src/master/src/Vector/BLF/tests/unittests/events_from_binlog/
    """

    def _setup_instance(self):
        super()._setup_instance_helper(
            can.BLFWriter,
            can.BLFReader,
            binary_file=True,
            check_fd=True,
            check_comments=False,
            test_append=True,
            allowed_timestamp_delta=1.0e-6,
            preserves_channel=False,
            adds_default_channel=0,
        )

    def _read_log_file(self, filename):
        logfile = os.path.join(os.path.dirname(__file__), "data", filename)
        with can.BLFReader(logfile) as reader:
            return list(reader)

    def test_can_message(self):
        expected = can.Message(
            timestamp=2459565876.494607,
            arbitration_id=0x4444444,
            is_extended_id=False,
            channel=0x1110,
            dlc=0x33,
            data=[0x55, 0x66, 0x77, 0x88, 0x99, 0xAA, 0xBB, 0xCC],
        )
        actual = self._read_log_file("test_CanMessage.blf")
        self.assertMessagesEqual(actual, [expected] * 2)
        self.assertEqual(actual[0].channel, expected.channel)

    def test_can_message_2(self):
        expected = can.Message(
            timestamp=2459565876.494607,
            arbitration_id=0x4444444,
            is_extended_id=False,
            channel=0x1110,
            dlc=0x33,
            data=[0x55, 0x66, 0x77, 0x88, 0x99, 0xAA, 0xBB, 0xCC],
        )
        actual = self._read_log_file("test_CanMessage2.blf")
        self.assertMessagesEqual(actual, [expected] * 2)
        self.assertEqual(actual[0].channel, expected.channel)

    def test_can_fd_message(self):
        expected = can.Message(
            timestamp=2459565876.494607,
            arbitration_id=0x4444444,
            is_extended_id=False,
            channel=0x1110,
            dlc=64,
            is_fd=True,
            bitrate_switch=True,
            error_state_indicator=True,
            data=range(64),
        )
        actual = self._read_log_file("test_CanFdMessage.blf")
        self.assertMessagesEqual(actual, [expected] * 2)
        self.assertEqual(actual[0].channel, expected.channel)

    def test_can_fd_message_64(self):
        expected = can.Message(
            timestamp=2459565876.494607,
            arbitration_id=0x15555555,
            is_extended_id=False,
            is_remote_frame=True,
            channel=0x10,
            dlc=64,
            is_fd=True,
            is_rx=False,
            bitrate_switch=True,
            error_state_indicator=True,
        )
        actual = self._read_log_file("test_CanFdMessage64.blf")
        self.assertMessagesEqual(actual, [expected] * 2)
        self.assertEqual(actual[0].channel, expected.channel)

    def test_can_error_frame_ext(self):
        expected = can.Message(
            timestamp=2459565876.494607,
            is_error_frame=True,
            arbitration_id=0x19999999,
            is_extended_id=True,
            channel=0x1110,
            dlc=0x66,
            data=[0xCC, 0xDD, 0xEE, 0xFF, 0x11, 0x22, 0x33, 0x44],
        )
        actual = self._read_log_file("test_CanErrorFrameExt.blf")
        self.assertMessagesEqual(actual, [expected] * 2)
        self.assertEqual(actual[0].channel, expected.channel)

    def test_timestamp_to_systemtime(self):
        self.assertAlmostEqual(
            1636485425.999,
            blf.systemtime_to_timestamp(blf.timestamp_to_systemtime(1636485425.998908)),
            places=3,
        )
        self.assertAlmostEqual(
            1636485426.0,
            blf.systemtime_to_timestamp(blf.timestamp_to_systemtime(1636485425.999908)),
            places=3,
        )


class TestCanutilsFileFormat(ReaderWriterTest):
    """Tests can.CanutilsLogWriter and can.CanutilsLogReader"""

    def _setup_instance(self):
        super()._setup_instance_helper(
            can.CanutilsLogWriter,
            can.CanutilsLogReader,
            check_fd=True,
            test_append=True,
            check_comments=False,
            preserves_channel=False,
            adds_default_channel="vcan0",
        )


class TestCsvFileFormat(ReaderWriterTest):
    """Tests can.CSVWriter and can.CSVReader"""

    def _setup_instance(self):
        super()._setup_instance_helper(
            can.CSVWriter,
            can.CSVReader,
            check_fd=False,
            test_append=True,
            check_comments=False,
            preserves_channel=False,
            adds_default_channel=None,
        )


@unittest.skipIf(asammdf is None, "MF4 is unavailable")
class TestMF4FileFormat(ReaderWriterTest):
    """Tests can.MF4Writer and can.MF4Reader"""

    def _setup_instance(self):
        super()._setup_instance_helper(
            can.MF4Writer,
            can.MF4Reader,
            binary_file=True,
            check_comments=False,
            preserves_channel=False,
            allowed_timestamp_delta=1e-4,
            adds_default_channel=0,
        )


class TestSqliteDatabaseFormat(ReaderWriterTest):
    """Tests can.SqliteWriter and can.SqliteReader"""

    def _setup_instance(self):
        super()._setup_instance_helper(
            can.SqliteWriter,
            can.SqliteReader,
            check_fd=False,
            test_append=True,
            check_comments=False,
            preserves_channel=False,
            adds_default_channel=None,
        )

    @unittest.skip("not implemented")
    def test_file_like_explicit_stop(self):
        pass

    @unittest.skip("not implemented")
    def test_file_like_context_manager(self):
        pass

    def test_read_all(self):
        """
        testing :meth:`can.SqliteReader.read_all` with context manager and path-like object
        """
        # create writer
        print("writing all messages/comments")
        with self.writer_constructor(self.test_file_name) as writer:
            self._write_all(writer)

        # read all written messages
        print("reading all messages")
        with self.reader_constructor(self.test_file_name) as reader:
            read_messages = list(reader.read_all())

        # check if at least the number of messages matches;
        self.assertEqual(
            len(read_messages),
            len(self.original_messages),
            "the number of written messages does not match the number of read messages",
        )

        self.assertMessagesEqual(self.original_messages, read_messages)


class TestPrinter(unittest.TestCase):
    """Tests that can.Printer does not crash.

    TODO test append mode
    """

    messages = (
        TEST_MESSAGES_BASE
        + TEST_MESSAGES_REMOTE_FRAMES
        + TEST_MESSAGES_ERROR_FRAMES
        + TEST_MESSAGES_CAN_FD
    )

    def test_not_crashes_with_stdout(self):
        with can.Printer() as printer:
            for message in self.messages:
                printer(message)

    def test_not_crashes_with_file(self):
        with tempfile.NamedTemporaryFile("w", delete=False) as temp_file:
            with can.Printer(temp_file) as printer:
                for message in self.messages:
                    printer(message)


class TestTrcFileFormatBase(ReaderWriterTest):
    """
    Base class for Tests with can.TRCWriter and can.TRCReader

    .. note::
        This class is prevented from being executed as a test
        case itself by a *del* statement in at the end of the file.
    """

    def _setup_instance(self):
        super()._setup_instance_helper(
            can.TRCWriter,
            can.TRCReader,
            check_remote_frames=False,
            check_error_frames=False,
            check_fd=False,
            check_comments=False,
            preserves_channel=False,
            allowed_timestamp_delta=0.001,
            adds_default_channel=0,
        )

    def _read_log_file(self, filename, **kwargs):
        logfile = os.path.join(os.path.dirname(__file__), "data", filename)
        with can.TRCReader(logfile, **kwargs) as reader:
            return list(reader)


class TestTrcFileFormatGen(TestTrcFileFormatBase):
    """Generic tests for can.TRCWriter and can.TRCReader with different file versions"""

    def test_can_message(self):
        start_time = 1506809173.191  # 30.09.2017 22:06:13.191.000 as timestamp
        expected_messages = [
            can.Message(
                timestamp=start_time + 2.5010,
                arbitration_id=0xC8,
                is_extended_id=False,
                is_rx=False,
                channel=1,
                dlc=8,
                data=[9, 8, 7, 6, 5, 4, 3, 2],
            ),
            can.Message(
                timestamp=start_time + 17.876708,
                arbitration_id=0x6F9,
                is_extended_id=False,
                channel=0,
                dlc=0x8,
                data=[5, 0xC, 0, 0, 0, 0, 0, 0],
            ),
        ]
        actual = self._read_log_file("test_CanMessage.trc")
        self.assertMessagesEqual(actual, expected_messages)

    @parameterized.expand(
        [
            ("V1_0", "test_CanMessage_V1_0_BUS1.trc", False),
            ("V1_1", "test_CanMessage_V1_1.trc", True),
            ("V2_1", "test_CanMessage_V2_1.trc", True),
        ]
    )
    def test_can_message_versions(self, name, filename, is_rx_support):
        with self.subTest(name):
            if name == "V1_0":
                # Version 1.0 does not support start time
                start_time = 0
            else:
                start_time = (
                    1639837687.062001  # 18.12.2021 14:28:07.062.001 as timestamp
                )

            def msg_std(timestamp):
                msg = can.Message(
                    timestamp=timestamp + start_time,
                    arbitration_id=0x000,
                    is_extended_id=False,
                    channel=1,
                    dlc=8,
                    data=[0, 0, 0, 0, 0, 0, 0, 0],
                )
                if is_rx_support:
                    msg.is_rx = False
                return msg

            def msg_ext(timestamp):
                msg = can.Message(
                    timestamp=timestamp + start_time,
                    arbitration_id=0x100,
                    is_extended_id=True,
                    channel=1,
                    dlc=8,
                    data=[0, 0, 0, 0, 0, 0, 0, 0],
                )
                if is_rx_support:
                    msg.is_rx = False
                return msg

            expected_messages = [
                msg_ext(17.5354),
                msg_ext(17.7003),
                msg_ext(17.8738),
                msg_std(19.2954),
                msg_std(19.5006),
                msg_std(19.7052),
                msg_ext(20.5927),
                msg_ext(20.7986),
                msg_ext(20.9560),
                msg_ext(21.0971),
            ]
            actual = self._read_log_file(filename)
            self.assertMessagesEqual(actual, expected_messages)

    def test_not_supported_version(self):
        with self.assertRaises(NotImplementedError):
            writer = can.TRCWriter("test.trc")
            writer.file_version = can.TRCFileVersion.UNKNOWN
            writer.on_message_received(can.Message())


class TestTrcFileFormatV1_0(TestTrcFileFormatBase):
    """Tests can.TRCWriter and can.TRCReader with file version 1.0"""

    @staticmethod
    def Writer(filename):
        writer = can.TRCWriter(filename)
        writer.file_version = can.TRCFileVersion.V1_0
        return writer

    def _setup_instance(self):
        super()._setup_instance()
        self.writer_constructor = TestTrcFileFormatV1_0.Writer


# this excludes the base class from being executed as a test case itself
del ReaderWriterTest
del TestTrcFileFormatBase


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python


import logging
import random
import threading
import unittest

logging.getLogger(__file__).setLevel(logging.WARNING)

# make a random bool:
rbool = lambda: bool(round(random.random()))

import can

channel = "vcan0"


class ControllerAreaNetworkTestCase(unittest.TestCase):
    """
    This test ensures that what messages go in to the bus is what comes out.

    Requires a can interface.

    To ensure that hardware and/or software message priority queues don't
    effect the test, messages are sent one at a time.
    """

    num_messages = 512

    # TODO check if error flags are working (don't currently appear on bus)
    error_flags = [False for _ in range(num_messages)]

    remote_flags = [rbool() for _ in range(num_messages)]
    extended_flags = [rbool() for _ in range(num_messages)]

    ids = list(range(num_messages))
    data = list(
        bytearray([random.randrange(0, 2**8 - 1) for a in range(random.randrange(9))])
        for b in range(num_messages)
    )

    def setUp(self):
        # Save all can.rc defaults
        self._can_rc = can.rc
        can.rc = {"interface": "virtual"}

    def tearDown(self):
        # Restore the defaults
        can.rc = self._can_rc

    def producer(self, ready_event, msg_read):
        self.client_bus = can.interface.Bus(channel=channel)
        ready_event.wait()
        for i in range(self.num_messages):
            m = can.Message(
                arbitration_id=self.ids[i],
                is_remote_frame=self.remote_flags[i],
                is_error_frame=self.error_flags[i],
                is_extended_id=self.extended_flags[i],
                data=self.data[i],
            )
            # logging.debug("writing message: {}".format(m))
            if msg_read is not None:
                # Don't send until the other thread is ready
                msg_read.wait()
                msg_read.clear()

            self.client_bus.send(m)

    def testProducer(self):
        """Verify that we can send arbitrary messages on the bus"""
        logging.debug("testing producer alone")
        ready = threading.Event()
        ready.set()
        self.producer(ready, None)

        logging.debug("producer test complete")

    def testProducerConsumer(self):
        logging.debug("testing producer/consumer")
        ready = threading.Event()
        msg_read = threading.Event()

        self.server_bus = can.interface.Bus(channel=channel)

        t = threading.Thread(target=self.producer, args=(ready, msg_read))
        t.start()

        # Ensure there are no messages on the bus
        while True:
            m = self.server_bus.recv(timeout=0.5)
            if m is None:
                print("No messages... lets go")
                break
            else:
                self.fail("received messages before the test has started ...")
        ready.set()
        i = 0
        while i < self.num_messages:
            msg_read.set()
            msg = self.server_bus.recv(timeout=0.5)
            self.assertIsNotNone(msg, "Didn't receive a message")
            # logging.debug("Received message {} with data: {}".format(i, msg.data))

            self.assertEqual(msg.is_extended_id, self.extended_flags[i])
            if not msg.is_remote_frame:
                self.assertEqual(msg.data, self.data[i])
            self.assertEqual(msg.arbitration_id, self.ids[i])

            self.assertEqual(msg.is_error_frame, self.error_flags[i])
            self.assertEqual(msg.is_remote_frame, self.remote_flags[i])

            i += 1
        t.join()

        self.server_bus.flush_tx_buffer()
        self.server_bus.shutdown()


if __name__ == "__main__":
    unittest.main()
